# /// script
# requires-python = ">=3.13"
# dependencies = [
#     "marimo",
#     "aignostics==0.2.57",
# ]
# ///

import marimo

__generated_with = "0.13.0"
app = marimo.App(width="full")


@app.cell
def _():
    import marimo as mo
    return (mo,)


@app.cell
def _(mo):
    mo.md(
        r"""
        # Initialize the Client

        As a first step, you need to initialize the client to interact with the Aignostics Platform. This will execute an OAuth flow depending on the environment you run:
        - In case you have a browser available, an interactive login flow in your browser is started.
        - In case there is no browser available, a device flow is started.

        **NOTE:** By default, the client caches the access token in your operation systems application cache folder. If you do not want to store the access token, please initialize the client like this:

        ```python
        import aignostics.client as platform
        # initialize the client
        client = platform.Client(cache_token=False)
        ```
        """
    )
    return


@app.cell
def _():
    import pandas as pd
    from pydantic import BaseModel

    # the following function is used for visualizing the results nicely in this notebook
    def show(models: BaseModel | list[BaseModel]) -> pd.DataFrame:
        if isinstance(models, BaseModel):
            items = [models.model_dump()]
        else:
            items = (a.model_dump() for a in models)
        return pd.DataFrame(items)
    return (show,)


@app.cell
def _():
    from aignostics import platform
    # initialize the client
    client = platform.Client(cache_token=False)
    return client, platform


@app.cell
def _(mo):
    mo.md(
        r"""
        # List our available applications

        Next, let us list the applications that are available in your organization:
        """
    )
    return


@app.cell
def _(client, show):
    applications = client.applications.list()
    # visualize
    show(applications)
    return


@app.cell
def _(mo):
    mo.md(
        r"""
        # List all available versions of an application

        Now that we know the applications that are available, we can list all the versions of a specific application. In this case, we will use the `TwoTask Dummy Application` as an example, which has the `application_id`: `two-task-dummy`. Using the `application_id`, we can list all the versions of the application:
        """
    )
    return


@app.cell
def _(client, show):
    application_versions = client.applications.versions.list(application="two-task-dummy")
    # visualize
    show(application_versions)
    return


@app.cell
def _(mo):
    mo.md(
        r"""
        # Inspect the application version details

        Now that we have the list of versions, we can inspect the details of a specific version. While we could directly use the list of application version returned by the `list` method, we want to directly query details for a specific application version. In this case, we will use version `0.35.0`, which has the `application_version_id`: `two-task-dummy:v0.35.0`. We use the `application_version_id` to retrieve further details about the application version:
        """
    )
    return


@app.cell
def _(client):
    two_task_app = client.applications.versions.details(application_version="two-task-dummy:v0.35.0")

    # view the `input_artifacts` to get insights in the required fields of the application version payload
    two_task_app.input_artifacts[0].to_json()
    return


@app.cell
def _(mo):
    mo.md(
        r"""
        # Trigger an application run

        Now, let's trigger an application run for the `Test Application`. We will use the `application_version_id` that we retrieved in the previous step. To create an application run, we need to provide a payload that consists of 1 or more items. We provide the Pydantic model `InputItem` an item and the data that comes with it:
        ```python
        platform.InputItem(
            reference="<a unique reference associate outputs to this input item>",
            input_artifacts=[platform.InputArtifact]
        )
        ```
        The `InputArtifact` defines the actual data that you provide aka. in this case the image that you want to be processed. The expected values are defined by the application version and have to align with the `input_artifacts` schema of the application version. In the case of the two task dummy application, we only require a single artifact per item, which is the image to process on. The artifact name is defined as `user_slide`. The `download_url` is a signed URL that allows the Aignostics Platform to download the image data later during processing. In addition to the image data itself, you have to provide the metadata defined in the input artifact schema, i.e., `checksum_crc32c`, `base_mpp`, `width`, and `height`. The metadata is used to validate the input data and is required for the processing of the image. The following example shows how to create an item with a single input artifact:

        ```python
        platform.InputArtifact(
            name="user_slide", # as defined by the application version input_artifact schema
            download_url="<a signed url to download the data>",
            metadata={
                "checksum_crc32c": "<checksum>",
                "base_mpp": "<base_mpp>",
                "width": "<width>",
                "height": "<height>"
            }
        )
        ```
        """
    )
    return


@app.cell
def _(client, platform):
    application_run = client.runs.create(
        application_version="two-task-dummy:v0.0.5",
        items=[
            platform.InputItem(
                reference="wsi-1",
                input_artifacts=[
                    platform.InputArtifact(
                        name="user_slide",
                        download_url=platform.generate_signed_url("<signed-url>"),
                        metadata={
                            "checksum_crc32c": "AAAAAA==",
                            "base_mpp": 0.25,
                            "width": 10000,
                            "height": 10000,
                        },
                    )
                ],
            ),
        ],
    )
    print(application_run)
    return (application_run,)


@app.cell
def _(mo):
    mo.md(
        r"""
        # Observe the status of the application run and download

        While you can observe the status of an application run directly via the `status()` method and also retrieve the results via the `results()` method, you can also download the results directly to a folder of your choice. The `download_to_folder()` method will download all the results to the specified folder. The method will automatically create a sub-folder in the specified folder with the name of the application run. The results for each individual input item will be stored in a separate folder named after the `reference` you defined in the `Item`.

        The method downloads the results for a slide as soon as they are available. There is no need to keep the method running until all results are available. The method will automatically check for the status of the application run and download the results as soon as they are available. If you invoke the method on a run you already downloaded some results before, it will only download the missing artifacts.
        """
    )
    return


@app.cell
def _(application_run):
    download_folder = "/tmp/"
    application_run.download_to_folder(download_folder)
    return


@app.cell
def _(mo):
    mo.md(
        r"""
        # Continue to retrieve results for an application run

        In case you just triggered an application run and want to check on the results later or you had a connection loss, you can simply initialize an applicaiton run object via it's `application_run_id`. If you do not have the `application_run_id` anymore, you can simple list all currently running application version via the `client.runs.list()` method. The `application_run_id` is part of the `ApplicationRun` object returned by the `list()` method. You can then use the `download_to_folder()` method to continue downloading the results.
        """
    )
    return


@app.cell
def _(client):
    # list currently running applications
    application_runs = client.runs.list()
    for run in application_runs:
        print(run)
    return


@app.cell
def _(mo):
    mo.md(
        r"""
        from aignostics.client.resources.runs import ApplicationRun
        application_run = ApplicationRun.for_application_run_id("<application_run_id>")
        # download
        download_folder = "/tmp/"
        application_run.download_to_folder(download_folder)
        """
    )
    return


if __name__ == "__main__":
    app.run()



"""Settings of the QuPath module."""

from pydantic_settings import SettingsConfigDict

from ..utils import OpaqueSettings, __env_file__, __project_name__  # noqa: TID252


class Settings(OpaqueSettings):
    """Settings."""

    model_config = SettingsConfigDict(
        env_prefix=f"{__project_name__.upper()}_QUPATH_",
        extra="ignore",
        env_file=__env_file__,
        env_file_encoding="utf-8",
    )



"""Theming."""

from pathlib import Path

from aignostics.utils import BasePageBuilder


class PageBuilder(BasePageBuilder):
    @staticmethod
    def register_pages() -> None:
        from nicegui import app  # noq  # noqa: PLC0415

        assets = Path(__file__).parent / "assets"
        app.add_static_files("/assets", assets)


def theme() -> None:
    """Theming incl. colors and fonts."""
    from nicegui import app, ui  # noqa: PLC0415

    ui.colors(
        primary="#1C1242",
        secondary="#B9B1DF",
        accent="#111B1E",
        dark="#1d1d1d1d",
        dark_page="#12121212",
        positive="#0CA57B",
        negative="#D4313C",
        info="#261C8D",
        warning="#FFCC00",
        brand_white="#EFF0F1",
        brand_background_light="#E7E6E8",
        brand_logo="#AFA3DD",
    )

    ui.add_head_html("""
        <style type="text/tailwindcss">
            @layer components {
                .blue-box {
                    @apply bg-blue-500 p-12 text-center shadow-lg rounded-lg text-white;
                }
            }
            @font-face{
                font-family: "cabin";
                src: url('/assets/cabin-v27-latin-regular.woff2') format('woff2');
                font-weight: normal;
                font-style: normal;
            }
            ::-webkit-scrollbar {
                display: none;
            }
            .bg-warning {
                color: black !important;
            }
            .bg-aignostics-light {
                background-color: #ECEDE9 !important;
            }
            .bg-aignostics-dark {
                background-color: #000000 !important;
            }
            .q-stepper, .q-drawer {
                background-color: #F0F0F0 !important;
            }
            .q-drawer.q-dark {
                background-color: #000000 !important;
            }
            html *
            {
                font-family: "Cabin";
            }
            header {
                color: white
            }
            footer {
                padding-top:0px; padding-left: 0px; height: 30px; background-color: white
            }
            .nicegui-markdown {
                ol {
                    padding-left: 20px;
                }
            }
        </style>
    """)

    ui.add_body_html(
        '<script src="https://unpkg.com/@dotlottie/player-component@2.7.12/dist/'
        'dotlottie-player.mjs" type="module"></script>'
    )

    ui.dark_mode(app.storage.general.get("dark_mode", False))



"""Exceptions of system module."""


class OpenAPISchemaError(ValueError):
    """Exception raised when OpenAPI schema cannot be loaded."""

    def __init__(self, error: Exception) -> None:
        """Initialize exception with the underlying error."""
        super().__init__(f"Failed to load OpenAPI schema: {error}")



"""Utilities module."""

from ._cli import prepare_cli
from ._console import console
from ._constants import (
    __author_email__,
    __author_name__,
    __base__url__,
    __documentation__url__,
    __env__,
    __env_file__,
    __is_development_mode__,
    __is_running_in_container__,
    __is_running_in_read_only_environment__,
    __project_name__,
    __project_path__,
    __repository_url__,
    __version__,
)
from ._di import load_modules, locate_implementations, locate_subclasses
from ._fs import get_user_data_directory, open_user_data_directory, sanitize_path, sanitize_path_component
from ._health import Health
from ._log import LogSettings, get_logger
from ._process import ProcessInfo, get_process_info
from ._service import BaseService
from ._settings import UNHIDE_SENSITIVE_INFO, OpaqueSettings, load_settings, strip_to_none_before_validator
from .boot import boot

__all__ = [
    "UNHIDE_SENSITIVE_INFO",
    "BaseService",
    "Health",
    "LogSettings",
    "OpaqueSettings",
    "ProcessInfo",
    "__author_email__",
    "__author_name__",
    "__base__url__",
    "__documentation__url__",
    "__env__",
    "__env_file__",
    "__is_development_mode__",
    "__is_running_in_container__",
    "__is_running_in_read_only_environment__",
    "__project_name__",
    "__project_path__",
    "__repository_url__",
    "__version__",
    "boot",
    "console",
    "get_logger",
    "get_process_info",
    "get_user_data_directory",
    "load_modules",
    "load_settings",
    "locate_implementations",
    "locate_subclasses",
    "open_user_data_directory",
    "prepare_cli",
    "sanitize_path",
    "sanitize_path_component",
    "strip_to_none_before_validator",
]

from importlib.util import find_spec

if find_spec("logfire"):
    from ._logfire import LogfireSettings, logfire_initialize

    __all__ += ["LogfireSettings", "logfire_initialize"]

if find_spec("sentry"):
    from ._sentry import SentrySettings

    __all__ += ["SentrySettings"]

if find_spec("nicegui"):
    from ._gui import BasePageBuilder, GUILocalFilePicker, gui_register_pages, gui_run

    __all__ += ["BasePageBuilder", "GUILocalFilePicker", "gui_register_pages", "gui_run"]

if find_spec("marimo"):
    from ._notebook import create_marimo_app

    __all__ += [
        "create_marimo_app",
    ]



"""Service of the QuPath module."""

import contextlib
import json
import platform
import queue
import re
import shutil
import subprocess
import tarfile
import tempfile
import time
import zipfile
from collections.abc import Callable
from enum import StrEnum
from pathlib import Path
from typing import Any
from urllib.parse import urlparse, urlsplit

import appdirs
import ijson
import psutil
import requests
from packaging.version import Version
from psutil import Process, wait_procs
from pydantic import BaseModel, computed_field

from aignostics.utils import UNHIDE_SENSITIVE_INFO, BaseService, Health, __project_name__, get_logger

from ._settings import Settings

logger = get_logger(__name__)

QUPATH_VERSION = "0.6.0-rc5"
DOWNLOAD_CHUNK_SIZE = 10 * 1024 * 1024
QUPATH_LAUNCH_MAX_WAIT_TIME = 30  # seconds, maximum wait time for QuPath to start
QUPATH_SCRIPT_MAX_EXECUTION_TIME = 60 * 60 * 2  # seconds, maximum wait time for QuPath to run a script

PROJECT_FILENAME = "project.qpproj"
ANNOTATIONS_BATCH_SIZE = 500000


class QuPathVersion(BaseModel):
    """Class to store QuPath version information."""

    version: str
    build_time: str | None = None
    commit_tag: str | None = None


class InstallProgressState(StrEnum):
    """Enum for download progress states."""

    CHECKING = "Trying to find QuPath ..."
    DOWNLOADING = "Downloading QuPath archive ..."
    EXTRACTING = "Extracting QuPath archive ..."


class InstallProgress(BaseModel):
    status: InstallProgressState = InstallProgressState.CHECKING
    archive_version: str | None = None
    archive_path: Path | None = None
    archive_size: int | None = None
    archive_downloaded_size: int = 0
    archive_download_chunk_size: int | None = None

    @computed_field  # type: ignore
    @property
    def archive_download_progress_normalized(self) -> float:
        """Compute normalized archive download progress in range 0..1.

        Returns:
            float: The normalized archive download progress in range 0..1.
        """
        if (not self.archive_size) or self.archive_size is None:
            return 0.0
        return min(1, float(self.archive_downloaded_size + 1) / float(self.archive_size))


class AddProgressState(StrEnum):
    """Enum for download progress states."""

    INITIALIZING = "Initializing ..."
    ADDING_IMAGES = "Adding images ..."
    COMPLETED = "Completed."


class AddProgress(BaseModel):
    status: AddProgressState = AddProgressState.INITIALIZING
    image_count: int | None = None
    image_index: int = 0
    image_path: Path | None = None

    @computed_field  # type: ignore
    @property
    def progress_normalized(self) -> float:
        """Compute normalized progress in range 0..1.

        Returns:
            float: The normalized progress in range 0..1.
        """
        if not self.image_count:
            return 0.0
        return min(1, float(self.image_index + 1) / float(self.image_count))


class AnnotateProgressState(StrEnum):
    """Enum for download progress states."""

    INITIALIZING = "Initializing ..."
    COUNTING = "Counting annotations ..."
    ANNOTATING = "Annotating image ..."
    COMPLETED = "Completed."


class AnnotateProgress(BaseModel):
    status: AnnotateProgressState = AnnotateProgressState.INITIALIZING
    image_path: Path | None = None
    annotation_count: int | None = None
    annotation_index: int = 0
    annotation_path: Path | None = None

    @computed_field  # type: ignore
    @property
    def progress_normalized(self) -> float:
        """Compute normalized progress in range 0..1.

        Returns:
            float: The normalized progress in range 0..1.
        """
        if not self.annotation_count:
            return 0.0
        return min(1, float(self.annotation_index) / float(self.annotation_count))


class QuPathImageHierarchy(BaseModel):
    """Hierarchy information for a QuPath image."""

    total: int
    detections: int
    annotations: int
    has_root_object: bool


class QuPathImage(BaseModel):
    """Information about a single image in a QuPath project."""

    id: str
    name: str
    description: str

    unique_name: str
    original_image_name: str
    uris: list[str]

    @computed_field  # type: ignore
    @property
    def path(self) -> str | None:
        """Get the path to the image data.

        Iterates over URIs and uses the first file URI found to extract the filename.

        Returns:
            Path: The path to the image data.
        """
        for uri in self.uris:
            parsed_uri = urlparse(uri)
            if parsed_uri.scheme == "file":
                return str(Path(parsed_uri.path).resolve())
        return None

    height: int
    width: int
    downsample_levels: list[float]

    num_channels: int
    num_timepoints: int
    num_zslices: int

    hierarchy: QuPathImageHierarchy

    entry_path: str
    thumbnail_path: str
    server_path: str
    data_path: str
    server_type: str
    server_builder: str


class QuPathProject(BaseModel):
    """Information about a QuPath project."""

    uri: str
    version: str
    images: list[QuPathImage]


class Service(BaseService):
    """Service of the bucket module."""

    _settings: Settings

    def __init__(self) -> None:
        """Initialize service."""
        super().__init__(Settings)

    def info(self, mask_secrets: bool = True) -> dict[str, Any]:
        """Determine info of this service.

        Args:
            mask_secrets (bool): Whether to mask sensitive information in the output.

        Returns:
            dict[str,Any]: The info of this service.
        """
        settings = self._settings.model_dump(context={UNHIDE_SENSITIVE_INFO: not mask_secrets})
        executable = Service.find_qupath_executable()
        version = Service.get_version()
        return {
            "app": {
                "path": str(executable) if executable else None,
                "version": dict(version) if version else None,
                "expected_version": Service.get_expected_version(),
            },
            "settings": settings,
        }

    def health(self) -> Health:
        """Determine health of this service.

        Returns:
            Health: The health of the service.
        """
        return Health(
            status=Health.Code.UP,
            components={
                "application": self._determine_application_health(),
            },
        )

    @staticmethod
    def _determine_application_health() -> Health:
        """Determine we can reach a well known and secure endpoint.

        - Checks if health endpoint is reachable and returns 200 OK
        - Uses requests library for a direct connection check without authentication

        Returns:
            Health: The healthiness of the network connection via basic unauthenticated request.
        """
        try:
            version = Service.get_version()
            if not version:
                message = "QuPath not installed."
                return Health(status=Health.Code.DOWN, reason=message)
            if version.version != Service.get_expected_version():
                message = f"QuPath version mismatch: expected {QUPATH_VERSION}, got {version.version}"
                logger.warning(message)
                return Health(status=Health.Code.DOWN, reason=message)
        except Exception as e:
            message = f"Exception while checking health of QuPath application {e!s}"
            logger.exception(message)
            return Health(status=Health.Code.DOWN, reason=message)
        return Health(status=Health.Code.UP)

    @staticmethod
    def _get_app_dir_from_qupath_dir(qupath_dir: Path, platform_system: str | None) -> Path:
        """Get the QuPath application directory based on the platform system.

        Args:
            qupath_dir (Path): The QuPath installation directory.
            platform_system (str | None): The system platform (e.g., "Linux", "Darwin", "Windows").
                Will be auto-determined if not given

        Returns:
            str: The path to the QuPath application directory.

        Raises:
            FileNotFoundError: If the QuPath application directory is not a directory.
        """
        platform_system = platform_system or platform.system()
        platform_paths = {"Linux": "lib/app", "Darwin": "Contents/app", "Windows": "app"}
        app_dir = qupath_dir / platform_paths[platform_system]

        if not app_dir.is_dir():
            message = f"QuPath installation directory is not a directory: s{app_dir!s}"
            raise FileNotFoundError(message)

        return app_dir

    @staticmethod
    def is_installed(
        platform_system: str | None = None,
    ) -> bool:
        """Check if QuPath is installed.

        Args:
            platform_system (str | None): The system platform. If None, it will use platform.system().

        Returns:
            bool: True if QuPath is installed, False otherwise.
        """
        return Service.get_version(platform_system=platform_system) is not None

    @staticmethod
    def get_expected_version() -> str:
        """Get expected version.

        Returns:
            QuPathVersion | None: The version of QuPath if installed, otherwise None.
        """
        return QUPATH_VERSION

    @staticmethod
    def get_version(platform_system: str | None = None) -> QuPathVersion | None:
        """Get the version of the installed QuPath.

        Args:
            platform_system (str | None): The system platform. If None, it will use platform.system().

        Returns:
            QuPathVersion | None: The version of QuPath if installed, otherwise None.
        """
        exectuable = Service.find_qupath_executable(platform_system=platform_system)

        if not exectuable:
            return None

        try:
            result = subprocess.run(  # noqa: S603
                [str(exectuable), "--version"],
                capture_output=True,
                text=True,
                check=True,
                timeout=QUPATH_LAUNCH_MAX_WAIT_TIME,
            )

            output = result.stdout.strip()
            logger.debug("QuPath version output: %s", output)

            version_match = re.search(r"Version:\s+([0-9]+\.[0-9]+\.[0-9]+(?:-rc[0-9]+)?)", output)

            # If standard pattern fails, try to match "QuPath vX.X.X" pattern, for Win32
            if not version_match:
                version_match = re.search(r"QuPath\s+v([0-9]+\.[0-9]+\.[0-9]+(?:-rc[0-9]+)?)", output)

            build_time_match = re.search(r"Build time:\s+(.+)", output)
            commit_tag_match = re.search(r"Latest commit tag:\s+[\"']?(.+?)[\"']?(?:\s|$)", output)

            if version_match:
                version = version_match.group(1)
                build_time = build_time_match.group(1) if build_time_match else None
                commit_tag = commit_tag_match.group(1) if commit_tag_match else None

                return QuPathVersion(version=version, build_time=build_time, commit_tag=commit_tag)
        except (subprocess.SubprocessError, subprocess.TimeoutExpired) as e:
            logger.warning("Failed to get QuPath version from executable: %s", e)
        except Exception:
            logger.exception("Unexpected error getting QuPath version")

        return None

    @staticmethod
    def _find_qupath_app_dir(platform_system: str | None) -> Path | None:
        """Find QuPath application directory.

        Args:
            platform_system (str | None): The system platform.

        Returns:
            Path | None: Path to QuPath app directory if found, None otherwise.
        """
        for d in Service.get_installation_path().iterdir():
            if d.is_dir() and re.match(r"(?i)qupath.*", d.name):
                try:
                    return Service._get_app_dir_from_qupath_dir(d, platform_system)
                except FileNotFoundError:
                    continue
        return None

    @staticmethod
    def _find_qupath_exe_from_app_dir(app_dir: Path, system: str) -> Path | None:
        """Get QuPath executable from app directory based on platform.

        Args:
            app_dir (Path): Path to QuPath app directory.
            system (str): The system platform.

        Returns:
            Path | None: Path to QuPath executable if found, None otherwise.
        """
        try:
            if system == "Linux":
                return next(app_dir.parent.parent.joinpath("bin").glob("QuPath*"))
            if system == "Darwin":
                return next(app_dir.parent.joinpath("MacOS").glob("QuPath*"))
            if system == "Windows":
                app_exes = list(app_dir.parent.glob("QuPath*.exe"))
                if len(app_exes) != 2:  # noqa: PLR2004
                    logger.warning(
                        "Expected to find exactly 2 QuPath executables, got %s. "
                        "Please ensure you have the correct QuPath installation.",
                        app_exes,
                    )
                    return None
                return next(qp for qp in app_exes if "console" in qp.stem)
        except StopIteration:
            return None
        return None

    @staticmethod
    def find_qupath_executable(platform_system: str | None = None) -> Path | None:
        """Find path to QuPath executable.

        Args:
            platform_system (str | None): The system platform. If None, it will use platform.system().

        Returns:
            Path | None: Path to the QuPath executable if found, otherwise None.
        """
        system = platform_system or platform.system()
        app_dir = Service._find_qupath_app_dir(platform_system)

        if app_dir is None:
            return None

        app_exe = Service._find_qupath_exe_from_app_dir(app_dir, system)

        if not app_exe or not app_exe.is_file():
            if app_exe:
                logger.warning("Expected to find file at %s.", app_exe)
            return None

        return app_exe

    @staticmethod
    def is_qupath_installed() -> bool:
        """Check if QuPath is installed.

        Returns:
            bool: True if QuPath is installed, False otherwise.
        """
        return Service.find_qupath_executable() is not None

    @staticmethod
    def get_installation_path() -> Path:
        """Get the installation directory of QuPath.

        Returns:
            Path: The directory QuPath will be installed into.
        """
        return Path(appdirs.user_data_dir(__project_name__)).resolve()

    @staticmethod
    def _download_qupath(  # noqa: C901, PLR0912, PLR0913, PLR0915, PLR0917
        version: str,
        path: Path,
        platform_system: str | None = None,
        platform_machine: str | None = None,
        download_progress: Callable | None = None,  # type: ignore[type-arg]
        install_progress_queue: queue.Queue[InstallProgress] | None = None,
    ) -> Path:
        """Download QuPath from GitHub.

        Args:
            version (str): Version of QuPath to download.
            path (Path): Path to directory save the downloaded file to.
            platform_system (str | None): The system platform. If None, it will use platform.system().
            platform_machine  (str | None): The machine architecture. If None, it will use platform.machine().
            download_progress (Callable | None): Callback function for download progress.
            install_progress_queue (Any | None): Queue for download progress updates, if applicable.

        Raises:
            ValueError: If the platform.system() is not supported.
            RuntimeError: If the download fails or if the file cannot be saved.
            Exception: If there is an error during the download.

        Returns:
            Path: The path object of the downloaded file.
        """
        system = platform.system() if platform_system is None else platform_system
        machine = platform.machine() if platform_machine is None else platform_machine
        logger.debug("Downloading QuPath version %s for system %s and machine %s", version, system, machine)

        if system == "Linux":
            sys = "Linux"
            ext = "tar.xz"
        elif system == "Darwin":
            sys = "Mac"
            ext = "pkg"
        elif system == "Windows":
            sys = "Windows"
            ext = "zip"
        else:
            error_message = f"unsupported platform.system() == {system!r}"
            raise ValueError(error_message)

        if not version.startswith("v"):
            version = f"v{version}"

        if Version(version) > Version("0.4.4"):
            if system == "Darwin":
                sys = "Mac-arm64" if machine == "arm64" else "Mac-x64"
            name = f"QuPath-{version}-{sys}"
        elif Version(version) > Version("0.3.2"):
            if system == "Darwin":
                sys = "Mac-arm64" if machine == "arm64" else "Mac"
            name = f"QuPath-{version[1:]}-{sys}"
        elif "rc" not in version:
            name = f"QuPath-{version[1:]}-{sys}"
        else:
            name = f"QuPath-{version[1:]}"

        url = f"https://github.com/qupath/qupath/releases/download/{version}/{name}.{ext}"

        logger.debug("Downloading QuPath from %s", url)

        filename = Path(urlsplit(url).path).name
        filepath = path / filename

        try:  # noqa: PLR1702
            with requests.get(url, stream=True, timeout=60) as stream:
                stream.raise_for_status()
                download_size = int(stream.headers.get("content-length", 0))
                downloaded_size = 0
                with open(filepath, mode="wb") as file:
                    for chunk in stream.iter_content(chunk_size=DOWNLOAD_CHUNK_SIZE):
                        if chunk:
                            downloaded_size += len(chunk)
                            file.write(chunk)
                            if download_progress:
                                download_progress(filepath, download_size, len(chunk))
                            if install_progress_queue:
                                progress = InstallProgress(
                                    status=InstallProgressState.DOWNLOADING,
                                    archive_version=version,
                                    archive_path=filepath,
                                    archive_size=download_size,
                                    archive_downloaded_size=downloaded_size,
                                    archive_download_chunk_size=len(chunk),
                                )
                                install_progress_queue.put_nowait(progress)
            logger.debug("Downloaded QuPath archive to '%s'", filepath)
        except requests.RequestException as e:
            message = f"Failed to download QuPath from {url}="
            logger.exception(message)
            raise RuntimeError(message) from e
        except Exception:
            logger.exception("Error downloading QuPath from %s", url)
            with contextlib.suppress(OSError):
                filepath.unlink(missing_ok=True)
            raise
        else:
            return filepath

    @staticmethod
    def get_app_dir(
        version: str, installation_path: Path, platform_system: str | None = None, platform_machine: str | None = None
    ) -> Path:
        """Get the the QuPath application directory.

        Args:
            version (str): Version of QuPath.
            installation_path (Path): Path to the installation directory.
            platform_system (str | None): The system platform. If None, it will use platform.system().
            platform_machine (str | None): The machine architecture. If None, it will use platform.machine().

        Returns:
            str: The version of QuPath extracted from the

        Raises:
            ValueError: If the version does not match the expected pattern or if the system is unsupported.
        """
        system = platform.system() if platform_system is None else platform_system
        machine = platform.machine() if platform_machine is None else platform_machine
        logger.debug(
            "Getting QuPath application directory for version '%s', installation path '%s' on system '%s'",
            version,
            installation_path,
            system,
        )

        m = re.match(
            r"v?(?P<version>[0-9]+[.][0-9]+[.][0-9]+(-rc[0-9]+|-m[0-9]+)?)",
            version,
        )
        if not m:
            message = f"version '{version}' does not match expected QuPath version pattern"
            logger.error(message)
            raise ValueError(message)
        version = m.group("version")

        if system == "Windows":
            return installation_path / Path(f"QuPath-{version}")
        if system == "Linux":
            return installation_path / Path("QuPath")
        if system == "Darwin":
            arch = "arm64" if machine == "arm64" else "x64"
            return installation_path / Path(f"QuPath-{version}-{arch}.app")
        message = f"unsupported platform.system() == {system!r}"
        raise ValueError(message)

    @staticmethod
    def _extract_qupath(  # noqa: C901, PLR0912, PLR0915
        archive_path: Path,
        installation_path: Path,
        overwrite: bool = False,
        platform_system: str | None = None,
        platform_machine: str | None = None,
    ) -> Path:
        """Extract downloaded QuPath installation archive to the specified destination directory.

        Args:
            archive_path (Path): Path to the downloaded QuPath archive.
            installation_path (Path): Path to the directory where QuPath should be extracted.
            overwrite (bool): If True, will overwrite existing files in the installation path.
            platform_system (str | None): The system platform. If None, it will use platform.system().
            platform_machine (str | None): The machine architecture. If None, it will use platform.machine().

        Raises:
            ValueError: If there is broken input.
            RuntimeError: If an unexpected error happens.

        Returns:
            Path: The path to the extracted QuPath application directory.
        """
        system = platform.system() if platform_system is None else platform_system
        logger.debug("Extracting QuPath archive '%s' to '%s' for system %s", archive_path, installation_path, system)

        destination = Service.get_app_dir(
            version=QUPATH_VERSION,
            installation_path=installation_path,
            platform_system=platform_system,
            platform_machine=platform_machine,
        )

        if destination.is_dir():
            if overwrite:
                with tempfile.TemporaryDirectory() as nirvana:
                    message = (
                        f"QuPath installation directory already exists at '{destination!s}', moving to nirvana ..."
                    )
                    logger.warning(message)
                    try:
                        shutil.move(destination, nirvana)
                    except PermissionError as e:
                        message = (
                            f"Failed to move existing QuPath directory '{destination!s}' "
                            f"to nirvana in '{nirvana!s}': {e!s}"
                        )
                        logger.exception(message)
                        raise RuntimeError(message) from e
            else:
                message = f"QuPath installation directory already exists at '{destination!s}', breaking. "
                logger.warning(message)
                return destination

        if system == "Linux":
            if not archive_path.name.endswith(".tar.xz"):
                message = f"archive '{archive_path!r}' does not end with `.tar.xz`"
                logger.error(message)
                raise ValueError(message)
            with tempfile.TemporaryDirectory() as tmp_dir:
                with tarfile.open(archive_path, mode="r:xz") as tf:
                    tf.extractall(tmp_dir)  # nosec: B202  # noqa: S202
                    for path in Path(tmp_dir).iterdir():
                        name = path.name
                        if name.startswith("QuPath") and path.is_dir():
                            break
                    else:
                        message = "No QuPath directory found in the extracted contents."
                        logger.error(message)
                        raise RuntimeError(message)
                extract_dir = Path(tmp_dir) / name
                if (extract_dir / "QuPath").is_dir():
                    # in some cases there is a nested QuPath directory
                    extract_dir /= "QuPath"
                shutil.move(extract_dir, installation_path)
            archive_path.unlink(missing_ok=True)  # remove the archive after extraction
            return destination

        if system == "Darwin":
            if archive_path.suffix != ".pkg":
                message = f"archive '{archive_path!s}' does not end with `.pkg`"
                logger.error(message)
                raise ValueError(message)

            if platform.system() not in {"Darwin", "Linux"}:
                message = f"Unsupported platform.system() == {platform.system()!r} for pkgutil"
                logger.error(message)
                raise ValueError(message)

            with tempfile.TemporaryDirectory() as tmp_dir:
                expanded_pkg_dir = Path(tmp_dir) / "expanded_pkg"  # pkgutil will create the directory
                try:
                    command = (
                        ["pkgutil", "--expand", str(archive_path.resolve()), str(expanded_pkg_dir.resolve())]
                        if platform.system() == "Darwin"
                        else ["7z", "x", str(archive_path.resolve()), "-o" + str(expanded_pkg_dir.resolve())]
                    )
                    subprocess.run(  # noqa: S603
                        command,
                        capture_output=True,
                        check=True,
                    )
                except subprocess.CalledProcessError as e:
                    stderr_output = e.stderr.decode("utf-8", errors="replace") if e.stderr else ""
                    message = f"Failed to expand .pkg file: {e!s}\nstderr:\n{stderr_output}"
                    logger.exception(message)
                    raise RuntimeError(message) from e

                payload_path = None
                for path in expanded_pkg_dir.rglob("Payload*"):
                    if path.is_file() and (path.name == "Payload" or path.name.startswith("Payload")):
                        payload_path = path
                        break
                if not payload_path:
                    message = "No Payload file found in the expanded .pkg"
                    logger.error(message)
                    raise RuntimeError(message)
                payload_extract_dir = Path(tmp_dir) / "payload_contents"
                payload_extract_dir.mkdir(parents=True, exist_ok=True)
                try:
                    command = (
                        [
                            "sh",
                            "-c",
                            f"cd '{payload_extract_dir.resolve()!s}' && "
                            f"cat '{payload_path.resolve()!s}' | gunzip -dc | cpio -i",
                        ]
                        if platform.system() == "Darwin"
                        else ["7z", "x", str(payload_path.resolve()), "-o" + str(payload_extract_dir.resolve())]
                    )
                    subprocess.run(  # noqa: S603
                        command,
                        capture_output=True,
                        check=True,
                    )
                except subprocess.CalledProcessError as e:
                    stderr_output = e.stderr.decode("utf-8", errors="replace") if e.stderr else ""
                    message = f"Failed to expand .pkg file: {e!s}\nstderr:\n{stderr_output}"
                    logger.exception(message)
                    raise RuntimeError(message) from e

                for app_path in payload_extract_dir.glob("**/*"):
                    if app_path.is_dir() and app_path.name.startswith("QuPath") and app_path.name.endswith(".app"):
                        shutil.move(app_path, installation_path)
                        archive_path.unlink(missing_ok=True)  # remove the archive after extraction
                        return destination

                message = "No QuPath application found in the extracted contents"
                logger.error(message)
                raise RuntimeError(message)

        if system == "Windows":
            if archive_path.suffix != ".zip":
                message = f"archive '{archive_path!s}' does not end with `.zip`"
                logger.error(message)
                raise ValueError(message)

            destination.parent.mkdir(parents=True, exist_ok=True)
            with zipfile.ZipFile(archive_path, mode="r") as zf:
                zf.extractall(destination)  # nosec: B202  # noqa: S202
                # Verify that QuPath executable exists in the extracted content
                qupath_exe_found = False
                for item in destination.rglob("*"):
                    if item.name.startswith("QuPath") and item.is_file() and item.suffix == ".exe":
                        qupath_exe_found = True
                        break
                if not qupath_exe_found:
                    message = "No QuPath .exe found in the extracted contents."
                    logger.error(message)
                    raise RuntimeError(message)
            archive_path.unlink(missing_ok=True)  # remove the archive after extraction
            return destination

        message = f"unsupported platform.system() == {system!r}"
        logger.error(message)
        raise RuntimeError(message)

    @staticmethod
    def install_qupath(  # noqa: PLR0913, PLR0917
        version: str = QUPATH_VERSION,
        path: Path | None = None,
        reinstall: bool = True,
        platform_system: str | None = None,
        platform_machine: str | None = None,
        download_progress: Callable | None = None,  # type: ignore[type-arg]
        extract_progress: Callable | None = None,  # type: ignore[type-arg]
        progress_queue: queue.Queue[InstallProgress] | None = None,
    ) -> Path:
        """Install QuPath application.

        Args:
            version (str): Version of QuPath to install. Defaults to "0.5.1".
            path (Path | None): Path to install QuPath to.
                If not specified, the home directory of the user will be used.
            reinstall (bool): If True, will reinstall QuPath even if it is already installed.
            platform_system (str | None): The system platform. If None, it will use platform.system().
            platform_machine (str | None): The machine architecture. If None, it will use platform.machine().
            download_progress (Callable | None): Callback function for download progress.
            extract_progress (Callable | None): Callback function for extraction progress.
            progress_queue (queue.Queue[InstallProgress] | None): Queue for download progress updates, if applicable.

        Raises:
            RuntimeError: If the download fails or if the file cannot be extracted.
            Exception: If there is an error during the download or extraction.

        Returns:
            Path: The path to the executable of the installed QuPath application.
        """
        if path is None:
            path = Service.get_installation_path()

        if not path.exists():
            try:
                path.mkdir(parents=True, exist_ok=True)
            except OSError as e:
                message = f"Failed to create installation directory '{path}': {e!s}"
                logger.exception(message)
                raise RuntimeError(message) from e
        try:
            archive_path = Service._download_qupath(
                version=version,
                path=path,
                platform_system=platform_system,
                platform_machine=platform_machine,
                download_progress=download_progress,
                install_progress_queue=progress_queue,
            )
            message = f"QuPath archive downloaded to '{archive_path!s}'."
            logger.debug(message)

            application_path = Service._extract_qupath(
                archive_path=archive_path,
                installation_path=path,
                overwrite=reinstall,
                platform_system=platform_system,
                platform_machine=platform_machine,
            )
            if not application_path.is_dir():
                message = f"QuPath directory not found as expected at '{application_path!s}'."
                logger.error(message)
                raise RuntimeError(message)  # noqa: TRY301
            message = f"QuPath application extracted to '{application_path!s}'."
            logger.debug(message)

            if extract_progress:
                application_size = 0
                for file_path in application_path.glob("**/*"):
                    if file_path.is_file():
                        application_size += file_path.stat().st_size
                message = f"Total size of QuPath application: '{application_size}' bytes"
                logger.debug(message)
                extract_progress(application_path, application_size=application_size)

            executable = Service.find_qupath_executable(
                platform_system=platform_system,
            )
            if not executable:
                message = "QuPath executable not found after installation."
                logger.error(message)
                raise RuntimeError(message)  # noqa: TRY301
            message = f"QuPath executable found at '{executable!s}'."
            logger.debug(message)

            executable.chmod(0o755)  # Make sure the executable is runnable
            message = f"Set permissions set to 755 for QuPath executable at '{executable!s}'."
            logger.debug(message)

            return application_path
        except Exception as e:
            message = f"Failed to install QuPath v{version} to '{path!s}': {e!s}"
            logger.exception(message)
            raise RuntimeError(message) from e

    @staticmethod
    def execute_qupath(  # noqa: C901, PLR0912, PLR0915, PLR0911
        quiet: bool = True,
        project: Path | None = None,
        image: str | Path | None = None,
        script: str | Path | None = None,
        script_args: list[str] | None = None,
    ) -> int | None:
        """Execute QuPath application.

        Args:
            quiet (bool): If True, will launch QuPath in quiet mode (no GUI).
            project (Path | None): Path to the QuPath project to open. If None, no project will be opened.
            image: str | Path | None: Path to the image file to open in QuPath. If project path given as well,
                this must be the name of the image within project as str.
            script (str | Path | None): Path to the script to run in QuPath. If None, no script will be run.
            script_args (list[str] | None): Arguments to pass to the script. Only used when script is provided.

        Returns:
            int | None: Pid if QuPath was launched successfully, None otherwise.

        """
        executable = Service.find_qupath_executable()
        if not executable:
            logger.error("QuPath executable not found.")
            return None

        message = f"QuPath executable found at: {executable}"
        logger.debug(message)

        if platform.system() in {"Linux", "Darwin", "Windows"}:
            command = [str(executable)]
            if script:
                command.extend(["script"])
            if quiet and not script:
                command.append("-q")
            if image:
                command.extend(["-i", str(image)])
            if project:
                command.extend(["-p", str(project.resolve() / PROJECT_FILENAME)])
            if script:
                command.extend([str(script)])
                if script_args:
                    for arg in script_args:
                        command.extend(["-a", arg])
        else:
            message = f"Unsupported platform: {platform.system()}"
            logger.error(message)
            raise NotImplementedError(message)

        try:
            logger.debug("Launching QuPath with command: %s", " ".join(command))
            process = subprocess.Popen(  # noqa: S603
                command,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                text=True,
                start_new_session=True,
            )
            if not process.stdout:
                logger.error("QuPath process has no stdout.")
                return None

            start_time = time.time()
            while True:
                waited = time.time() - start_time
                if script:
                    if waited > QUPATH_SCRIPT_MAX_EXECUTION_TIME:
                        message = f"Timed out after {waited:.2f} seconds waiting for QuPath script to complete"
                        logger.error(message)
                        return None
                elif waited > QUPATH_LAUNCH_MAX_WAIT_TIME:
                    message = f"Timed out after {waited:.2f} seconds waiting for QuPath to launch"
                    logger.error(message)
                    return None

                output = process.stdout.readline() if process.stdout.readable() else ""
                exit_code = process.poll()

                if not output and exit_code is not None:
                    if script:
                        if exit_code == 0:
                            logger.debug("QuPath script completed successfully.")
                            return process.pid
                        message = f"QuPath script failed with exit code '{exit_code}'."
                        logger.error(message)
                        return None
                    message = f"QuPath process has terminated with exit code '{exit_code}'."
                    logger.error(message)
                    break

                if output:
                    logger.debug(output.strip())
                    if "qupath.lib.gui.QuPathApp - Starting QuPath with parameters" in output:
                        logger.debug("QuPath started successfully.")
                        return process.pid

                # Small sleep to prevent CPU hogging
                time.sleep(0.1)
            return None
        except Exception as exc:
            message = f"Failed to launch QuPath: {exc!s}"
            logger.exception(message)
            return None

    @staticmethod
    def get_qupath_processes() -> list[Process]:
        """Get PIDs of QuPath processes started from our managed installation.

        Returns:
            list[Process]: List of processes handles of running QuPath instances.
                Processes not started from installation part are ignored.
        """
        installation_path = Service.get_installation_path()
        processes = []
        for proc in psutil.process_iter(attrs={"pid", "exe"}):
            try:
                exe_path = proc.exe()
                if exe_path and str(installation_path) in exe_path:
                    processes.append(Process(proc.pid))
            except (psutil.NoSuchProcess, psutil.AccessDenied):
                continue
        return processes

    @staticmethod
    def terminate_qupath_processes(wait_before_kill: int = 3) -> int:
        """Terminate QuPath processes.

        Args:
            wait_before_kill (int): Time in seconds to wait before forcefully killing processes.

        Returns:
            int: Number of processes terminated.

        """
        procs = Service.get_qupath_processes()
        for p in procs:
            try:
                p.terminate()
            except psutil.NoSuchProcess:
                logger.debug("Process %s already terminated.", p.pid)
                continue
        _, alive = wait_procs(procs, timeout=wait_before_kill)
        for p in alive:
            try:
                p.kill()
            except psutil.NoSuchProcess:
                logger.debug("Process %s already terminated.", p.pid)
                continue
        return len(procs)

    @staticmethod
    def uninstall_qupath(
        version: str | None = None,
        path: Path | None = None,
        platform_system: str | None = None,
        platform_machine: str | None = None,
    ) -> bool:
        """Uninstall QuPath application.

        Notice:
            Terminates managed QuPath processes before uninstalling.

        Args:
            version (str): Specific version of QuPath to uninstall. Defaults to None,
                i.e. all versions will be uninstalled.
            path (Path | None): Path to the directory where QuPath is installed.
                If not specified, the default installation path will be used.
            platform_system (str | None): The system platform. If None, it will use platform.system().
            platform_machine (str | None): The machine architecture. If None, it will use platform.machine().

        Returns:
            bool: True if QuPath was uninstalled successfully.

        Raises:
            ValueError: If the QuPath application directory is a file unexpectedly.
        """
        if path is None:
            path = Service.get_installation_path()

        Service.terminate_qupath_processes()

        if not version:
            removed = False
            for qupath in path.glob("QuPath*"):
                if qupath.is_dir():
                    logger.debug("Removing QuPath directory '%s'", qupath)
                else:
                    logger.debug("Removing QuPath archive '%s'", qupath)
                try:
                    shutil.rmtree(qupath, ignore_errors=False)
                    removed = True
                except Exception as e:  # noqa: BLE001
                    message = f"Failed to remove '{qupath!s}': {e!s}"
                    logger.warning(message)
            return removed

        app_dir = Service.get_app_dir(
            version=version,
            installation_path=path,
            platform_system=platform_system,
            platform_machine=platform_machine,
        )
        if not app_dir.exists():
            message = f"QuPath application directory '{app_dir!s}' does not exist, skipping."
            logger.warning(message)
            return False
        if app_dir.is_file():
            message = f"QuPath application directory '{app_dir!s}' is a file unexpectedly."
            logger.error(message)
            raise ValueError(message)
        try:
            logger.debug("Removing '%s'", app_dir)
            shutil.rmtree(app_dir, ignore_errors=False)
            return True
        except Exception as e:  # noqa: BLE001
            message = f"Failed to remove '{app_dir!s}': {e!s}"
            logger.warning(message)
        return False

    @staticmethod
    def _find_groovy_script(script_name: str) -> Path:
        """Find the Groovy script file in the scripts directory.

        Args:
            script_name (str): Name of the Groovy script file to find.

        Returns:
            Path: The path to the Groovy script file.

        Raises:
            ValueError: If the script file does not exist.
        """
        script_path = Path(__file__).parent / "scripts" / f"{script_name}.groovy"
        if not script_path.is_file():
            message = f"Groovy script '{script_name}' not found at: {script_path}"
            logger.error(message)
            raise ValueError(message)
        return script_path

    @staticmethod
    def add(
        project: Path,
        paths: list[Path],
        progress_callable: Callable | None = None,  # type: ignore[type-arg]
    ) -> int:
        """Add images to a QuPath project.

        Args:
            project (Path): Path to the QuPath project directory. Will be created if not existent.
            paths (list[Path]): One or multiple paths. A path can point to an individual image or folder.
                In case of a folder, all images within will be added for supported image types
            progress_callable (Callable | None): Optional callback function to report progress of adding images.

        Returns:
            int: The number of images added to the project.

        Raises:
            ValueError: If QuPath is not installed or the project path is invalid.
            RuntimeError: If there is an unexpected error adding images to the project.
        """
        message = f"Adding images to QuPath project at '{project!s}' from paths: {paths!r}"
        logger.debug(message)

        if progress_callable:
            progress = AddProgress()
            progress_callable(progress)

        if progress_callable:
            progress.status = AddProgressState.ADDING_IMAGES
            progress_callable(progress)

        # We communicate via file I/O with the Groovy script running within QuPath
        with (
            tempfile.NamedTemporaryFile(mode="w", suffix=".json", encoding="utf-8") as paths_file,
            tempfile.NamedTemporaryFile(mode="w", suffix=".json", encoding="utf-8") as output_file,
        ):
            try:
                json.dump([str(path.resolve()) for path in paths], paths_file.file)
                paths_file.flush()

                pid = Service.execute_qupath(
                    script=Service._find_groovy_script("add"),
                    script_args=[str(project), str(paths_file.name), str(output_file.name)],
                )

                if not pid:
                    message = "Failed to execute QuPath script for adding images."
                    logger.error(message)
                    raise RuntimeError(message)  # noqa: TRY301

                with Path(output_file.name).open("r", encoding="utf-8") as f:
                    result_data = json.load(f)
                added_count = int(result_data.get("added_count", 0))
                errors = result_data.get("errors", [])
                for error in errors:
                    logger.warning("QuPath add script error: %s", error)

                if progress_callable:
                    progress.status = AddProgressState.COMPLETED
                    progress_callable(progress)

                return added_count
            except Exception as e:
                message = f"Failed to add images to QuPath project: {e!s}"
                logger.exception(message)
                raise RuntimeError(message) from e

    @staticmethod
    def annotate(
        project: Path,
        image: Path,
        annotations: Path,
        progress_callable: Callable | None = None,  # type: ignore[type-arg]
    ) -> int:
        """Annotate an image in a QuPath project.

        Args:
            project (Path): Path to the QuPath project directory. Will be created if not existent.
            image (Path): Path to the image file to annotate. Will be added to the project if not already present.
            annotations (Path): Path to the annotations file in compatible GeoJSON format.
            progress_callable (Callable | None): Optional callback function to report progress of annotating the image.

        Returns:
            int: The number of annotations added to the image.

        Raises:
            ValueError: If QuPath is not installed, the project path is invalid or the annotation path is invalid.
            RuntimeError: If there is an error annotating the image.
        """
        if progress_callable:
            progress = AnnotateProgress()
            progress_callable(progress)

        if not image.is_file():
            message = f"Image path '{image!s}' is not a valid file."
            logger.error(message)
            raise ValueError(message)

        if not annotations.is_file():
            message = f"Annotations path '{annotations!s}' is not a valid file."
            logger.error(message)
            raise ValueError(message)

        if progress_callable:
            progress.annotation_path = annotations
            progress.image_path = image
            progress_callable(progress)

        if progress_callable:
            progress.status = AnnotateProgressState.COUNTING
            progress_callable(progress)

        annotation_count = 0
        with open(annotations, "rb") as f:
            features_parser = ijson.items(f, "features.item")
            for _ in features_parser:
                annotation_count += 1

        if progress_callable:
            progress.annotation_count = annotation_count
            progress.status = AnnotateProgressState.ANNOTATING
            progress_callable(progress)

        pid = Service.execute_qupath(
            project=project,
            script=Service._find_groovy_script("annotate"),
            script_args=[image.name, str(annotations.as_posix())],
        )
        if not pid:
            message = "Failed to execute QuPath script for annotations."
            logger.error(message)
            raise RuntimeError(message)

        if progress_callable:
            progress.status = AnnotateProgressState.COMPLETED
            progress_callable(progress)

        return annotation_count

    @staticmethod
    def inspect(project: Path) -> QuPathProject:
        """Inspect QuPath project using a Groovy script.

        Args:
            project (Path): Path to the QuPath project directory.

        Returns:
            QuPathProject: The QuPath project information.

        Raises:
            ValueError: If QuPath is not installed or the project path is invalid.
            RuntimeError: If there is an error inspecting the project.
        """
        # We communicate via file I/O with the Groovy script running within QuPath
        with tempfile.NamedTemporaryFile(mode="w+", suffix=".json", encoding="utf-8") as temp_file:
            output_file = Path(temp_file.name).resolve()

            pid = Service.execute_qupath(
                quiet=True,
                project=project,
                script=Service._find_groovy_script("inspect"),
                script_args=[str(output_file)],
            )

            if not pid:
                message = "Failed to execute QuPath inspect script"
                logger.error(message)
                raise RuntimeError(message)

            if not output_file.is_file():
                message = f"QuPath script output file '{output_file!s}' does not exist."
                logger.error(message)
                raise RuntimeError(message)

            return QuPathProject.model_validate_json(output_file.read_text(encoding="utf-8"))



"""Logfire integration for logging and instrumentation."""

import os
from importlib.util import find_spec
from typing import Annotated

from pydantic import BeforeValidator, Field, PlainSerializer, SecretStr
from pydantic_settings import SettingsConfigDict

from ._constants import __env__, __env_file__, __project_name__, __repository_url__, __version__
from ._settings import OpaqueSettings, load_settings, strip_to_none_before_validator


class LogfireSettings(OpaqueSettings):
    """Configuration settings for Logfire integration."""

    model_config = SettingsConfigDict(
        env_prefix=f"{__project_name__.upper()}_LOGFIRE_",
        env_file=__env_file__,
        env_file_encoding="utf-8",
        extra="ignore",
    )

    enabled: Annotated[
        bool,
        Field(
            description="Enable remote log collection via logfire",
            default=False,
        ),
    ]

    token: Annotated[
        SecretStr | None,
        BeforeValidator(strip_to_none_before_validator),
        PlainSerializer(func=OpaqueSettings.serialize_sensitive_info, return_type=str, when_used="always"),
        Field(description="Logfire token. Leave empty to disable logfire.", examples=["YOUR_TOKEN"], default=None),
    ]

    instrument_system_metrics: Annotated[
        bool,
        Field(description="Enable system metrics instrumentation", default=False),
    ]


def logfire_initialize(modules: list["str"]) -> bool:
    """Initialize Logfire integration.

    Args:
        modules(list["str"]): List of modules to be instrumented.

    Returns:
        bool: True if initialized successfully False otherwise
    """
    settings = load_settings(LogfireSettings)

    if not find_spec("logfire") or not settings.enabled or settings.token is None:
        os.environ["LOGFIRE_PYDANTIC_RECORD"] = "off"
        return False

    import logfire  # noqa: PLC0415

    logfire.configure(
        send_to_logfire="if-token-present",
        token=settings.token.get_secret_value(),
        environment=__env__,
        service_name=__project_name__,
        console=False,
        code_source=logfire.CodeSource(
            repository=__repository_url__,
            revision=__version__,
            root_path="",
        ),
    )

    if settings.instrument_system_metrics:
        logfire.instrument_system_metrics(base="full")

    logfire.instrument_pydantic()

    logfire.install_auto_tracing(modules=modules, min_duration=0.0)

    return True



"""Service of dataset module."""

import atexit
import re
import subprocess
import sys
import threading
import time
from multiprocessing import Queue
from pathlib import Path
from typing import Any

from aignostics.utils import BaseService, Health, get_logger

logger = get_logger(__name__)

PATH_LENGTH_MAX = 260
TARGET_LAYOUT_DEFAULT = "%collection_id/%PatientID/%StudyInstanceUID/%Modality_%SeriesInstanceUID/"

# Global registry of active processes for cleanup
_active_processes: list[subprocess.Popen[str]] = []


def _terminate_process(process: subprocess.Popen[str]) -> None:
    """Terminate a single subprocess with graceful shutdown attempt.

    Args:
        process: The subprocess to terminate.
    """
    try:
        logger.warning("Terminating orphaned subprocess with PID %d", process.pid)
        process.terminate()
        # Give it a moment to terminate gracefully
        for _ in range(5):
            if process.poll() is not None:
                break
            time.sleep(0.1)
        # If still running, force kill
        if process.poll() is None:
            logger.warning("Forcefully killing subprocess with PID %d", process.pid)
            process.kill()
    except Exception:
        message = f"Error terminating subprocess with PID {process.pid}"
        logger.exception(message)


def _cleanup_processes() -> None:
    """Terminate any active subprocesses on exit."""
    for process in _active_processes[:]:
        if process.poll() is None:  # Process is still running
            _terminate_process(process)
            _active_processes.remove(process)


# Register the cleanup function
atexit.register(_cleanup_processes)


class Service(BaseService):
    """Service of the IDC module."""

    def info(self, mask_secrets: bool = True) -> dict[str, Any]:  # noqa: ARG002, PLR6301
        """Determine info of this service.

        Args:
            mask_secrets (bool): Whether to mask sensitive information in the output.

        Returns:
            dict[str,Any]: The info of this service.
        """
        return {}

    def health(self) -> Health:  # noqa: PLR6301
        """Determine health of hello service.

        Returns:
            Health: The health of the service.
        """
        return Health(
            status=Health.Code.UP,
            components={},
        )

    @staticmethod
    def _capture_progress_output(  # noqa: C901
        process: subprocess.Popen[str],
        queue: Queue,  # type: ignore[type-arg]
        base_progress: float = 0.04,
    ) -> None:
        """Capture output from the download process and update progress queue.

        Args:
            process (subprocess.Popen): Process with stdout to monitor
            queue (Queue): Queue to update with progress information
            base_progress (float): Starting progress value (0.5 means 50% complete)
        """
        if process.stderr is None:
            logger.warning("Cannot capture progress: subprocess stderr is None")
            return

        progress_pattern = re.compile(r"Downloading data:\s+(\d+)%")
        buffer = ""
        last_percentage = 0

        # Read one character at a time to handle carriage returns
        while process.poll() is None:
            char = process.stderr.read(1)
            if not char:  # End of stream
                break

            char_str = char
            # Handle carriage return (line overwrite)
            if char_str == "\r":
                # Process the current buffer for percentage
                match = progress_pattern.search(buffer)
                if match:
                    percentage = int(match.group(1))
                    if percentage != last_percentage:  # Only update if changed
                        last_percentage = percentage
                        # Scale the progress
                        adjusted_progress = base_progress + (percentage / 100.0) * (1.0 - base_progress)
                        queue.put_nowait(min(adjusted_progress, 0.99))  # Cap at 99% until complete
                        logger.debug("Updated progress: %.2f", adjusted_progress)

                # Reset buffer after processing carriage return
                buffer = ""
            elif char_str == "\n":
                # Process the current buffer for percentage on newline
                match = progress_pattern.search(buffer)
                if match:
                    percentage = int(match.group(1))
                    if percentage != last_percentage:
                        last_percentage = percentage
                        adjusted_progress = base_progress + (percentage / 100.0) * (1.0 - base_progress)
                        queue.put_nowait(min(adjusted_progress, 0.99))
                        logger.debug("Updated progress: %.2f", adjusted_progress)

                # For debug purposes, log the complete line
                logger.debug("Process output: %s", buffer)
                buffer = ""
            else:
                # Add character to buffer
                buffer += char_str

        # Process any remaining content in buffer
        if buffer:
            match = progress_pattern.search(buffer)
            if match:
                percentage = int(match.group(1))
                adjusted_progress = base_progress + (percentage / 100.0) * (1.0 - base_progress)
                queue.put_nowait(min(adjusted_progress, 0.99))

        # Process has finished, set progress to 100%
        queue.put_nowait(1.0)
        logger.debug("Process completed, setting progress to 100%")

    @staticmethod
    def download_with_queue(  # noqa: PLR0915
        queue: Queue,  # type: ignore[type-arg]
        source: str,
        target: str = str(Path.cwd()),
        target_layout: str = TARGET_LAYOUT_DEFAULT,
        dry_run: bool = False,
    ) -> None:
        """Download from manifest file, identifier, or comma-separate set of identifiers.

        Args:
            queue (Queue): The queue to use for progress updates.
            source (str): The source to download from.
            target (str): The target directory to download to.
            target_layout (str): The layout of the target directory.
            dry_run (bool): If True, perform a dry run.

        Raises:
            ValueError: If the target directory does not exist.
        """
        from aignostics.third_party.idc_index import IDCClient  # noqa: PLC0415

        queue.put_nowait(0.01)

        client = IDCClient.client()
        queue.put_nowait(0.02)

        target_directory = Path(target)
        if not target_directory.is_dir():
            logger.warning("Target directory does not exist: %s", target_directory)
            message = f"Target directory does not exist: {target_directory}"
            raise ValueError(message)

        item_ids = [item.strip() for item in source.split(",") if item.strip()]

        if not item_ids:
            message = "No IDs provided."
            logger.warning(message)
            raise ValueError(message)

        index_df = client.index
        client.fetch_index("sm_instance_index")
        logger.info("Downloaded instance index")
        sm_instance_index_df = client.sm_instance_index
        queue.put_nowait(0.03)

        def check_and_download(column_name: str, item_ids: list[str], target_directory: Path, kwarg_name: str) -> bool:
            if column_name != "SOPInstanceUID":
                matches = index_df[column_name].isin(item_ids)
                matched_ids = index_df[column_name][matches].unique().tolist()
            else:
                matches = sm_instance_index_df[column_name].isin(item_ids)  # type: ignore
                matched_ids = sm_instance_index_df[column_name][matches].unique().tolist()  # type: ignore
            if not matched_ids:
                return False
            unmatched_ids = list(set(item_ids) - set(matched_ids))
            if unmatched_ids:
                logger.debug("Partial match for %s: matched %s, unmatched %s", column_name, matched_ids, unmatched_ids)
            logger.info("Identified matching %s: %s", column_name, matched_ids)
            queue.put_nowait(0.04)

            # Properly handle Windows paths - convert to raw string format
            safe_target_dir = str(target_directory).replace("\\", "\\\\")

            # Create command for the subprocess
            script_content = f"""
import sys
from aignostics.third_party.idc_index import IDCClient

client = IDCClient.client()
client.fetch_index("sm_instance_index")
client.download_from_selection(
    {kwarg_name}={matched_ids!r},
    downloadDir="{safe_target_dir}",
    dirTemplate="{target_layout}",
    quiet=False,
    show_progress_bar=True,
    use_s5cmd_sync=True,
    dry_run={dry_run!r}
)
"""

            # Run the download in a subprocess
            logger.debug(
                "Starting download subprocess with executable '%s' and script:\n%s", sys.executable, script_content
            )
            process = subprocess.Popen(  # noqa: S603
                [sys.executable, "-c", script_content],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
                bufsize=1,
            )

            # Register process for cleanup
            _active_processes.append(process)

            # Start a thread to monitor the subprocess output
            monitor_thread = threading.Thread(
                target=Service._capture_progress_output, args=(process, queue, 0.5), daemon=True
            )
            monitor_thread.start()

            try:
                # Wait for the subprocess to complete
                return_code = process.wait()
                monitor_thread.join()

                if return_code != 0:
                    stdout_output = process.stdout.read() if process.stdout else "No stdout output"
                    stderr_output = process.stderr.read() if process.stderr else "No stderr output"
                    logger.error(
                        "Download subprocess failed with code '%d'\n\nstdout:\n\n%sstdin:\n\n%s\n\n",
                        return_code,
                        stdout_output,
                        stderr_output,
                    )
                    return False

                logger.info("Download completed successfully")
                queue.put_nowait(1.0)
                return True
            finally:
                # Clean up process reference
                if process in _active_processes:
                    _active_processes.remove(process)

        matches_found = 0
        matches_found += check_and_download("collection_id", item_ids, target_directory, "collection_id")
        matches_found += check_and_download("PatientID", item_ids, target_directory, "patientId")
        matches_found += check_and_download("StudyInstanceUID", item_ids, target_directory, "studyInstanceUID")
        matches_found += check_and_download("SeriesInstanceUID", item_ids, target_directory, "seriesInstanceUID")
        matches_found += check_and_download("SOPInstanceUID", item_ids, target_directory, "sopInstanceUID")
        if not matches_found:
            message = "None of the values passed matched any of the identifiers: "
            message += "collection_id, PatientID, StudyInstanceUID, SeriesInstanceUID, SOPInstanceUID."
            logger.warning(message)
            raise ValueError(message)



"""Python SDK providing access to Aignostics AI services."""

import os

from .constants import MODULES_TO_INSTRUMENT, WSI_SUPPORTED_FILE_EXTENSIONS
from .utils.boot import boot

# Add scheme to HTTP proxy environment variables if missing
for proxy_var in ["HTTP_PROXY", "HTTPS_PROXY"]:
    proxy_url = os.environ.get(proxy_var)
    if proxy_url and not proxy_url.startswith(("http://", "https://")):
        os.environ[proxy_var] = f"http://{proxy_url}"

boot(modules_to_instrument=MODULES_TO_INSTRUMENT)

__all__ = ["WSI_SUPPORTED_FILE_EXTENSIONS"]



"""Notebook GUI."""

from pathlib import Path
from urllib.parse import quote

from aignostics.gui import frame, theme
from aignostics.utils import BasePageBuilder, get_logger, get_user_data_directory

logger = get_logger(__name__)


class PageBuilder(BasePageBuilder):
    @staticmethod
    def register_pages() -> None:
        from nicegui import app, ui  # noq  # noqa: PLC0415

        from ._service import Service  # noqa: PLC0415

        app.add_static_files("/notebook_assets", Path(__file__).parent / "assets")

        @ui.page("/notebook")
        def page_index() -> None:
            """Marimo Extension."""
            with frame("Marimo Extension", left_sidebar=False):
                # Nothing to do here, just to show the page
                pass

            def launch_marimo() -> None:
                """Launch QuPath."""
                launch_button.props(add="loading")
                ui.notify("Launching Python Notebook...", color="blue")
                ui.navigate.to(f"/notebook/all?results_folder={quote(get_user_data_directory('results').as_posix())}")

            ui.markdown(
                """
                    ### Manage your Marimo Extension
                """
            )
            with ui.row().classes("w-full justify-start"):
                with ui.column().classes("w-2/5"):
                    with ui.card().classes("w-full"):
                        ui.label(
                            "Marimo is installed and ready to execute. "
                            "Go to a completed application result and click the Notebook button."
                        )
                        with ui.row().classes("w-full justify-between items-center"):
                            launch_button = ui.button(
                                "Open",
                                on_click=launch_marimo,
                                icon="visibility",
                            ).mark("BUTTON_NOTEBOOK_LAUNCH")
                    ui.markdown(
                        """
                            ###### What is Marimo?
                            marimo [1, 2] is an open-source reactive notebook for Python — reproducible, git-friendly,
                            SQL built-in, executable as a script, and shareable as an app.

                            The Aignostics Launchpad embeds Marimo so you can apply data analysis on
                            your application results without having to leave the application.

                            *References:*

                            1. <a href="https://marimo.io/" target="_blank">Marimo website</a>

                            2. <a href="https://raw.githubusercontent.com/marimo-team/marimo/refs/heads/main/LICENSE"
                                target="_blank">License</a>
                        """
                    )
                ui.space()
                with ui.column().classes("w-2/5"), ui.row().classes("w-1/2 justify-center content-center"):
                    ui.space()
                    animation = "/notebook_assets/python.lottie"
                    ui.html(
                        f"<dotlottie-player "
                        f'src="{animation}" '
                        f'background="transparent" '
                        f'speed="1" '
                        f'style="width: 300px; height: 300px" '
                        f'direction="1" '
                        f'playMode="normal" '
                        f"loop "
                        f"autoplay>"
                        f"</dotlottie-player>"
                    )
                    ui.space()

        @ui.page("/notebook/{application_run_id}")
        def page_application_run_marimo(application_run_id: str, results_folder: str) -> None:
            """Inspect Application Run in Marimo."""
            theme()

            with ui.row().classes("w-full justify-end"):
                ui.button(
                    "Back to Application Run" if application_run_id != "all" else "Back to Marimo Extension",
                    icon="arrow_back",
                    on_click=lambda _, application_run_id=application_run_id: ui.navigate.to(
                        f"/application/run/{application_run_id}" if application_run_id != "all" else "/notebook"
                    ),
                ).mark("BUTTON_NOTEBOOK_BACK")

            try:
                server_url = Service().start()
                ui.html(
                    f'<iframe src="{server_url}?application_run_id={application_run_id}'
                    f'&results_folder={quote(results_folder)}" '
                    'width="100%" height="100%"></iframe>'
                ).classes("w-full h-[calc(100vh-5rem)]")
            except Exception:
                message = "Failed to start Marimo server."
                logger.exception("Failed to start Marimo server")
                ui.label(message).classes("text-red-500")
                ui.button("Retry", on_click=ui.navigate.reload).props("color=red")



"""Runs resource module for the Aignostics client.

This module provides classes for creating and managing application runs on the Aignostics platform.
It includes functionality for starting runs, monitoring status, and downloading results.
"""

import typing as t
from collections.abc import Generator
from pathlib import Path
from time import sleep
from typing import Any

from aignx.codegen.api.public_api import PublicApi
from aignx.codegen.models import (
    ApplicationRunStatus,
    ItemCreationRequest,
    ItemResultReadResponse,
    ItemStatus,
    RunCreationRequest,
    RunCreationResponse,
)
from aignx.codegen.models import (
    ItemResultReadResponse as ItemResultData,
)
from aignx.codegen.models import (
    RunReadResponse as ApplicationRunData,
)
from jsonschema.exceptions import ValidationError
from jsonschema.validators import validate

from aignostics.platform._utils import calculate_file_crc32c, download_file, mime_type_to_file_ending
from aignostics.platform.resources.applications import Versions
from aignostics.platform.resources.utils import paginate

LIST_APPLICATION_RUNS_MAX_PAGE_SIZE = 100
LIST_APPLICATION_RUNS_MIN_PAGE_SIZE = 5


class ApplicationRun:
    """Represents a single application run.

    Provides operations to check status, retrieve results, and download artifacts.
    """

    def __init__(self, api: PublicApi, application_run_id: str) -> None:
        """Initializes an ApplicationRun instance.

        Args:
            api (PublicApi): The configured API client.
            application_run_id (str): The ID of the application run.
        """
        self._api = api
        self.application_run_id = application_run_id

    @classmethod
    def for_application_run_id(cls, application_run_id: str) -> "ApplicationRun":
        """Creates an ApplicationRun instance for an existing run.

        Args:
            application_run_id (str): The ID of the application run.

        Returns:
            ApplicationRun: The initialized ApplicationRun instance.
        """
        from aignostics.platform._client import Client  # noqa: PLC0415

        return cls(Client.get_api_client(cache_token=False), application_run_id)

    # TODO(Andreas): Deprecated, please remove when you updated your integration code
    def status(self) -> ApplicationRunData:
        """Retrieves the current status of the application run.

        Returns:
            ApplicationRunData: The run data.

        Raises:
            Exception: If the API request fails.
        """
        return self.details()

    def details(self) -> ApplicationRunData:
        """Retrieves the current status of the application run.

        Returns:
            ApplicationRunData: The run data.

        Raises:
            Exception: If the API request fails.
        """
        return self._api.get_run_v1_runs_application_run_id_get(self.application_run_id, include=None)

    def item_status(self) -> dict[str, ItemStatus]:
        """Retrieves the status of all items in the run.

        Returns:
            dict[str, ItemStatus]: A dictionary mapping item references to their status.

        Raises:
            Exception: If the API request fails.
        """
        return {item.reference: item.status for item in self.results()}

    # TODO(Andreas): Fails with Internal Server Error if run canceled; don't throw generic exceptions
    def cancel(self) -> None:
        """Cancels the application run.

        Raises:
            Exception: If the API request fails.
        """
        self._api.cancel_application_run_v1_runs_application_run_id_cancel_post(self.application_run_id)

    def results(self) -> t.Iterator[ItemResultData]:
        """Retrieves the results of all items in the run.

        Returns:
            list[ItemResultData]: A list of item results.

        Raises:
            Exception: If the API request fails.
        """
        return paginate(
            self._api.list_run_results_v1_runs_application_run_id_results_get,
            application_run_id=self.application_run_id,
        )

    def download_to_folder(self, download_base: Path | str) -> None:
        """Downloads all result artifacts to a folder.

        Monitors run progress and downloads results as they become available.

        Args:
            download_base (Path | str): Base directory to download results to.

        Raises:
            ValueError: If the provided path is not a directory.
            Exception: If downloads or API requests fail.
        """
        # create application run base folder
        download_base = Path(download_base)
        if not download_base.is_dir():
            msg = f"{download_base} is not a directory"
            raise ValueError(msg)
        application_run_dir = Path(download_base) / self.application_run_id

        # incrementally check for available results
        application_run_status = self.details().status
        while application_run_status == ApplicationRunStatus.RUNNING:
            for item in self.results():
                if item.status == ItemStatus.SUCCEEDED:
                    self.ensure_artifacts_downloaded(application_run_dir, item)
            sleep(5)
            application_run_status = self.details().status
            print(self)

        # check if last results have been downloaded yet and report on errors
        for item in self.results():
            match item.status:
                case ItemStatus.SUCCEEDED:
                    self.ensure_artifacts_downloaded(application_run_dir, item)
                case ItemStatus.ERROR_SYSTEM | ItemStatus.ERROR_USER:
                    print(f"{item.reference} failed with {item.status.value}: {item.error}")

    @staticmethod
    def ensure_artifacts_downloaded(base_folder: Path, item: ItemResultReadResponse) -> None:
        """Ensures all artifacts for an item are downloaded.

        Downloads missing or partially downloaded artifacts and verifies their integrity.

        Args:
            base_folder (Path): Base directory to download artifacts to.
            item (ItemResultReadResponse): The item result containing the artifacts to download.

        Raises:
            ValueError: If checksums don't match.
            Exception: If downloads fail.
        """
        item_dir = base_folder / item.reference

        downloaded_at_least_one_artifact = False
        for artifact in item.output_artifacts:
            if artifact.download_url:
                item_dir.mkdir(exist_ok=True, parents=True)
                file_ending = mime_type_to_file_ending(artifact.mime_type)
                file_path = item_dir / f"{artifact.name}{file_ending}"
                checksum = artifact.metadata["checksum_crc32c"]

                if file_path.exists():
                    file_checksum = calculate_file_crc32c(file_path)
                    if file_checksum != checksum:
                        print(f"> Resume download for {artifact.name} to {file_path}")
                    else:
                        continue
                else:
                    downloaded_at_least_one_artifact = True
                    print(f"> Download for {artifact.name} to {file_path}")

                # if file is not there at all or only partially downloaded yet
                download_file(artifact.download_url, str(file_path), checksum)

        if downloaded_at_least_one_artifact:
            print(f"Downloaded results for item: {item.reference} to {item_dir}")
        else:
            print(f"Results for item: {item.reference} already present in {item_dir}")

    def __str__(self) -> str:
        """Returns a string representation of the application run.

        The string includes run ID, status, and item statistics.

        Returns:
            str: String representation of the application run.
        """
        app_status = self.details().status.value
        item_status = self.item_status()
        pending, succeeded, error = 0, 0, 0
        for item in item_status.values():
            match item:
                case ItemStatus.PENDING:
                    pending += 1
                case ItemStatus.SUCCEEDED:
                    succeeded += 1
                case ItemStatus.ERROR_USER | ItemStatus.ERROR_SYSTEM:
                    error += 1

        items = f"{len(item_status)} items - ({pending}/{succeeded}/{error}) [pending/succeeded/error]"
        return f"Application run `{self.application_run_id}`: {app_status}, {items}"


class Runs:
    """Resource class for managing application runs.

    Provides operations to create, find, and retrieve runs.
    """

    def __init__(self, api: PublicApi) -> None:
        """Initializes the Runs resource with the API client.

        Args:
            api (PublicApi): The configured API client.
        """
        self._api = api

    def __call__(self, application_run_id: str) -> ApplicationRun:
        """Retrieves an ApplicationRun instance for an existing run.

        Args:
            application_run_id (str): The ID of the application run.

        Returns:
            ApplicationRun: The initialized ApplicationRun instance.
        """
        return ApplicationRun(self._api, application_run_id)

    def create(self, application_version: str, items: list[ItemCreationRequest]) -> ApplicationRun:
        """Creates a new application run.

        Args:
            application_version (ApplicationVersion | str): The ID of the application version.
            items (list[ItemCreationRequest]): The run creation request payload.

        Returns:
            ApplicationRun: The created application run.

        Raises:
            ValueError: If the payload is invalid.
            Exception: If the API request fails.
        """
        payload = RunCreationRequest(
            application_version_id=application_version,
            items=items,
        )
        self._validate_input_items(payload)
        res: RunCreationResponse = self._api.create_application_run_v1_runs_post(payload)
        return ApplicationRun(self._api, str(res.application_run_id))

    def list(self, for_application_version: str | None = None) -> Generator[ApplicationRun, Any, None]:
        """Find application runs, optionally filtered by application version.

        Args:
            for_application_version (str | None): Optional application version ID to filter by.

        Returns:
            Generator[ApplicationRun, Any, None]: A generator yielding application runs.

        Raises:
            Exception: If the API request fails.
        """
        if not for_application_version:
            res = paginate(self._api.list_application_runs_v1_runs_get)
        else:
            res = paginate(self._api.list_application_runs_v1_runs_get, application_version_id=for_application_version)
        return (ApplicationRun(self._api, response.application_run_id) for response in res)

    # TODO(Andreas): Think about merging by having list(...) above return active records that as well hold data
    def list_data(
        self,
        for_application_version: str | None = None,
        sort: str | None = None,
        page_size: int = LIST_APPLICATION_RUNS_MAX_PAGE_SIZE,
    ) -> t.Iterator[ApplicationRunData]:
        """Fetch application runs, optionally filtered by application version.

        Args:
            for_application_version (str | None): Optional application version ID to filter by.
            sort (str | None): Optional field to sort by. Prefix with '-' for descending order.
            page_size (int): Number of items per page, defaults to max

        Returns:
            Iterator[ApplicationRunData]: Iterator yielding application run data.

        Raises:
            ValueError: If page_size is greater than 100.
            Exception: If the API request fails.
        """
        if page_size > LIST_APPLICATION_RUNS_MAX_PAGE_SIZE:
            message = (
                f"page_size is must be less than or equal to {LIST_APPLICATION_RUNS_MAX_PAGE_SIZE}, but got {page_size}"
            )
            raise ValueError(message)
        if not for_application_version:
            res = paginate(
                self._api.list_application_runs_v1_runs_get,
                page_size=page_size,
                sort=[sort] if sort else None,
            )
        else:
            res = paginate(
                self._api.list_application_runs_v1_runs_get,
                page_size=page_size,
                application_version_id=for_application_version,
                sort=[sort] if sort else None,
            )
        return res

    def _validate_input_items(self, payload: RunCreationRequest) -> None:
        """Validates the input items in a run creation request.

        Checks that references are unique, all required artifacts are provided,
        and artifact metadata matches the expected schema.

        Args:
            payload (RunCreationRequest): The run creation request payload.

        Raises:
            ValueError: If validation fails.
            Exception: If the API request fails.
        """
        # validate metadata based on schema of application version
        app_version = Versions(self._api).details(application_version=payload.application_version_id)
        schema_idx = {
            input_artifact.name: input_artifact.metadata_schema for input_artifact in app_version.input_artifacts
        }
        references = set()
        for item in payload.items:
            # verify references are unique
            if item.reference in references:
                msg = f"Duplicate reference `{item.reference}` in items."
                raise ValueError(msg)
            references.add(item.reference)

            schema_check = set(schema_idx.keys())
            for artifact in item.input_artifacts:
                # check if artifact is in schema
                if artifact.name not in schema_idx:
                    msg = f"Invalid artifact `{artifact.name}`, application version requires: {schema_idx.keys()}"
                    raise ValueError(msg)
                try:
                    # validate metadata
                    validate(artifact.metadata, schema=schema_idx[artifact.name])
                    schema_check.remove(artifact.name)
                except ValidationError as e:
                    msg = f"Invalid metadata for artifact `{artifact.name}`: {e.message}"
                    raise ValueError(msg) from e
            # all artifacts set?
            if len(schema_check) > 0:
                msg = f"Missing artifact(s): {schema_check}"
                raise ValueError(msg)



import platform
from abc import ABC, abstractmethod
from pathlib import Path
from types import EllipsisType

from ._constants import __is_running_in_container__, __project_name__
from ._di import locate_subclasses
from ._log import get_logger

logger = get_logger(__name__)

WINDOW_SIZE = (1200, 768)  # Default window size for the GUI


class BasePageBuilder(ABC):
    """Base class for all page builders."""

    @staticmethod
    @abstractmethod
    def register_pages() -> None:
        """Register pages."""


def gui_register_pages() -> None:
    """Register pages.

    This function is called by the GUI to register all pages.
    """
    page_builders = locate_subclasses(BasePageBuilder)
    for page_builder in page_builders:
        page_builder.register_pages()


def gui_run(  # noqa: PLR0913, PLR0917
    native: bool = True,
    show: bool = False,
    host: str | None = None,
    port: int | None = None,
    title: str = __project_name__,
    icon: str = "",
    watch: bool = False,
    with_api: bool = False,
    dark_mode: bool = False,
) -> None:
    """Start the GUI.

    Args:
        native: Whether to run the GUI in native mode.
        show: Whether to show the GUI.
        host: Host to run the GUI on.
        port: Port to run the GUI on.
        title: Title of the GUI.
        icon: Icon for the GUI.
        watch: Whether to watch for changes and reload the GUI.
        with_api: Whether to mount the API.
        dark_mode: Whether to use dark mode.

    Raises:
        ValueError: If with_notebook is True but notebook_path is None,
            or trying to run native within container.
    """
    from nicegui import native as native_app  # noqa: PLC0415
    from nicegui import ui  # noqa: PLC0415

    if __is_running_in_container__ and native:
        message = "Native GUI cannot be run in a container. Please run with uvx or in browser."
        raise ValueError(message)

    if with_api:
        message = "with_api is not supported in this project."
        raise ValueError(message)

    if native and platform.system() == "Linux":
        native = False
        show = True

    gui_register_pages()

    ui.run(
        title=title,
        favicon=icon,
        native=native,
        reload=watch,
        dark=dark_mode,
        host=host,
        port=port or native_app.find_open_port(),
        frameless=native and platform.system() == "Darwin",
        show_welcome_message=native is False,
        show=show,
        window_size=WINDOW_SIZE if native else None,
    )


class GUILocalFilePicker:
    """Local File Picker dialog class that lazy-loads NiceGUI dependencies."""

    def __new__(  # noqa: C901
        cls,
        directory: str,
        *,
        upper_limit: str | EllipsisType | None = ...,
        multiple: bool = False,
        show_hidden_files: bool = False,
    ) -> "GUILocalFilePicker":
        """Create a new instance with lazy-loaded dependencies.

        Args:
            directory: The directory to start in.
            upper_limit: The directory to stop at. None for no limit, default is same as starting directory.
            multiple: Whether to allow multiple files to be selected.
            show_hidden_files: Whether to show hidden files.

        Returns:
            An instance of the dialog with lazy-loaded dependencies.
        """
        from nicegui import app, events, ui  # noqa: PLC0415
        # Lazy import ui only when actually creating an instance

        # Define the actual implementation class with the imports available
        class GUILocalFilePickerImpl(ui.dialog):
            def __init__(
                self,
                directory: str,
                *,
                upper_limit: str | EllipsisType | None = ...,
                multiple: bool = False,
                show_hidden_files: bool = False,
            ) -> None:
                """Local File Picker.

                A simple file picker that allows selecting files from the local filesystem where NiceGUI is running.

                Args:
                    directory: The directory to start in.
                    upper_limit: The directory to stop at. None for no limit, default is same as starting directory.
                    multiple: Whether to allow multiple files to be selected.
                    show_hidden_files: Whether to show hidden files.
                """
                super().__init__()

                self.path = Path(directory).expanduser()
                if upper_limit is None:
                    self.upper_limit = None
                elif upper_limit is ...:
                    self.upper_limit = Path(directory).expanduser()
                else:
                    self.upper_limit = Path(upper_limit).expanduser()
                self.show_hidden_files = show_hidden_files

                with self, ui.card():
                    self.add_drives_toggle()
                    self.grid = (
                        ui.aggrid(
                            {
                                "columnDefs": [{"field": "name", "headerName": "File"}],
                                "rowSelection": "multiple" if multiple else "single",
                            },
                            html_columns=[0],
                        )
                        .classes("w-96")
                        .classes(
                            "ag-theme-balham-dark" if app.storage.general.get("dark_mode", False) else "ag-theme-balham"
                        )
                        .on("cellDoubleClicked", self.handle_double_click)
                    )
                    with ui.row().classes("w-full justify-end"):
                        ui.button("Cancel", on_click=self.close).props("outline").mark("BUTTON_FILEPICKER_CANCEL")
                        ui.button("Ok", on_click=self._handle_ok).mark("BUTTON_FILEPICKER_OK")
                self.update_grid()

            def add_drives_toggle(self) -> None:
                if platform.system() == "Windows":
                    import win32api  # type: ignore[unused-ignore] # type: ignore # noqa: PLC0415

                    drives = win32api.GetLogicalDriveStrings().split("\000")[:-1]
                    self.drives_toggle = ui.toggle(drives, value=drives[0], on_change=self.update_drive)

            def update_drive(self) -> None:
                self.path = Path(str(self.drives_toggle.value)).expanduser()
                self.update_grid()

            def update_grid(self) -> None:
                paths = list(self.path.glob("*"))
                if not self.show_hidden_files:
                    paths = [p for p in paths if not p.name.startswith(".")]
                paths.sort(key=lambda p: p.name.lower())
                paths.sort(key=lambda p: not p.is_dir())

                self.grid.options["rowData"] = [
                    {
                        "name": f"📁 <strong>{p.name}</strong>" if p.is_dir() else p.name,
                        "path": str(p),
                    }
                    for p in paths
                ]
                if (self.upper_limit is None and self.path != self.path.parent) or (
                    self.upper_limit is not None and self.path != self.upper_limit
                ):
                    self.grid.options["rowData"].insert(
                        0,
                        {
                            "name": "📁 <strong>..</strong>",
                            "path": str(self.path.parent),
                        },
                    )
                self.grid.update()

            def handle_double_click(self, e: events.GenericEventArguments) -> None:
                self.path = Path(e.args["data"]["path"])
                if self.path.is_dir():
                    self.update_grid()
                else:
                    self.submit([str(self.path)])

            async def _handle_ok(self) -> None:
                rows = await self.grid.get_selected_rows()
                self.submit([r["path"] for r in rows])

        # Create and return an instance but tell mypy it's a GUILocalFilePicker
        return GUILocalFilePickerImpl(  # type: ignore[return-value]
            directory=directory,
            upper_limit=upper_limit,
            multiple=multiple,
            show_hidden_files=show_hidden_files,
        )



"""CLI for operations on wsi files."""

from pathlib import Path
from typing import Annotated

import typer

from aignostics.utils import console, get_logger

from ._service import Service
from ._utils import print_slide_info, print_study_info

logger = get_logger(__name__)


cli = typer.Typer(name="wsi", help="Operations on whole slide images.")


@cli.command()
def inspect(
    path: Annotated[Path, typer.Argument(help="Path to the wsi file", exists=True)],
) -> None:
    """Inspect a wsi file and display its metadata."""
    metadata = Service().get_metadata(path)

    # Basics
    console.print("Format:", style="blue", end=" ")
    console.print(metadata["format"], style="green")
    console.print("Path:", style="blue", end=" ")
    console.print(metadata["file"]["path"], style="green")
    console.print("Size (human):", style="blue", end=" ")
    console.print(metadata["file"]["size_human"], style="green")
    console.print("Width:", style="blue", end=" ")
    console.print(metadata["dimensions"]["width"], style="green")
    console.print("Height:", style="blue", end=" ")
    console.print(metadata["dimensions"]["height"], style="green")
    console.print("MPP (x):", style="blue", end=" ")
    console.print(metadata["resolution"]["mpp_x"], style="green")
    console.print("MPP (y):", style="blue", end=" ")
    console.print(metadata["resolution"]["mpp_y"], style="green")

    # Image Properties
    if "properties" in metadata and "image" in metadata["properties"]:
        img = metadata["properties"]["image"]
        created = f"{img['date']} (libvips {img['version']})"
        console.print("Created:", style="blue", end=" ")
        console.print(created, style="green")

        if "properties" in img and "bands" in img["properties"]:
            console.print("Color channels:", style="blue", end=" ")
            console.print(str(img["properties"]["bands"]), style="green")

        if "properties" in img and "aix-original-format" in img["properties"]:
            console.print("aix-original-format:", style="blue", end=" ")
            console.print(str(img["properties"]["aix-original-format"]), style="green")

    # Level Structure
    console.print("\nLevel Structure:", style="bold blue")
    for level in metadata["levels"]["data"]:
        console.print(f"\nLevel {level['index']}", style="blue")

        dimensions = f"{level['dimensions']['width']} x {level['dimensions']['height']} pixels"
        console.print("  Dimensions:", style="blue", end=" ")
        console.print(dimensions, style="green")

        downsample = f"{level['downsample']:.1f}x"
        console.print("  Downsample factor:", style="blue", end=" ")
        console.print(downsample, style="green")

        pixel_size = f"{metadata['resolution']['mpp_x'] * level['downsample']:.3f} μm/pixel"
        console.print("  Pixel size:", style="blue", end=" ")
        console.print(pixel_size, style="green")

        tile_size = f"{level['tile']['width']} x {level['tile']['height']} pixels"
        console.print("  Tile size:", style="blue", end=" ")
        console.print(tile_size, style="green")

        tiles = f"{level['tile']['grid']['x']} x {level['tile']['grid']['y']} ({level['tile']['grid']['total']} total)"
        console.print("  Tiles:", style="blue", end=" ")
        console.print(tiles, style="green")

    # Associated Images
    if metadata.get("associated_images"):
        console.print("\nAssociated Images:", style="bold blue")
        for img in metadata["associated_images"]:
            console.print(f"  - {img}", style="green")


cli_dicom = typer.Typer(no_args_is_help=True)
cli.add_typer(cli_dicom, name="dicom")


@cli_dicom.command(name="inspect")
def dicom_inspect(
    path: Annotated[
        Path,
        typer.Argument(..., help="Path of file or directory to inspect", exists=True),
    ],
    verbose: Annotated[bool, typer.Option(help="Verbose output")] = False,
    summary: Annotated[bool, typer.Option(help="Show only summary information")] = False,
) -> None:  # pylint: disable=W0613
    """Inspect DICOM files at any hierarchy level."""
    from ._pydicom_handler import PydicomHandler  # noqa: PLC0415

    with PydicomHandler.from_file(str(path)) as handler:
        metadata = handler.get_metadata(verbose)

        if metadata["type"] == "empty":
            console.print("[bold red]No DICOM files found in the specified path.[/bold red]")
            return

        # Print hierarchy
        for study_uid, study_data in metadata["studies"].items():
            console.print(f"\n[bold]Study:[/bold] {study_uid}")
            print_study_info(study_data)

            if not summary:
                for container_id, slide_data in study_data["slides"].items():
                    console.print(f"\n[bold]Slide (Container ID):[/bold] {container_id}")
                    print_slide_info(slide_data, indent=1, verbose=verbose)


@cli_dicom.command(name="geojson_import")
def dicom_geojson_import(
    dicom_path: Annotated[Path, typer.Argument(help="Path to the DICOM file", exists=True)],
    geojson_path: Annotated[Path, typer.Argument(help="Path to the GeoJSON file", exists=True)],
) -> None:  # pylint: disable=W0613
    """Import GeoJSON annotations into DICOM ANN instance."""
    from ._pydicom_handler import PydicomHandler  # noqa: PLC0415

    console.print("\nImporting GeoJSON annotations into DICOM ANN instance...", style="blue")
    PydicomHandler.geojson_import(dicom_path, geojson_path)



"""Service of the platform module."""

import time
from datetime import datetime
from http import HTTPStatus
from typing import Any

import urllib3
from pydantic import BaseModel, computed_field

from aignostics.utils import UNHIDE_SENSITIVE_INFO, BaseService, Health, __version__, get_logger

from ._authentication import CLAIM_ROLE, get_token, remove_cached_token, userinfo, verify_and_decode_token
from ._client import Client
from ._settings import Settings

logger = get_logger(__name__)


class TokenInfo(BaseModel):
    """Class to store token information."""

    issuer: str  # iss
    issued_at: int  # iat
    expires_at: int  # exp
    scope: list[str]  # scope
    audience: list[str]  # aud
    authorized_party: str  # azp

    @computed_field  # type: ignore[prop-decorator]
    @property
    def expires_in(self) -> int:
        """Calculate seconds until token expires.

        Returns:
            int: Number of seconds until the token expires. Negative if already expired.
        """
        return self.expires_at - int(time.time())

    @classmethod
    def from_claims(cls, claims: dict[str, Any]) -> "TokenInfo":
        """Create TokenInfo from JWT claims.

        Args:
            claims: JWT token claims dictionary.

        Returns:
            TokenInfo: Token information extracted from claims.
        """
        return cls(
            issuer=claims["iss"],
            issued_at=claims["iat"],
            expires_at=claims["exp"],
            scope=claims["scope"].split(),
            audience=claims["aud"] if isinstance(claims["aud"], list) else [claims["aud"]],
            authorized_party=claims["azp"],
        )


class UserProfile(BaseModel):
    """Class to store info about the user."""

    name: str | None = None  # userinfo.name
    given_name: str | None = None  # userinfo.given_name
    family_name: str | None = None  # userinfo.family_name
    nickname: str | None = None  # userinfo.nickname
    email: str | None = None  # userinfo.email
    email_verified: bool | None = None  # userinfo.email_verified
    picture: str | None = None  # userinfo.picture
    updated_at: datetime | None = None  # userinfo.updated_at

    @classmethod
    def from_userinfo(cls, userinfo: dict[str, Any]) -> "UserProfile":
        """Create UserProfile from auth0 userinfo.

        Args:
            userinfo (dict[str, Any] | None): User information dictionary from auth0.

        Returns:
            UserProfile: User information extracted from auth0 userinfo.
        """
        return cls(
            name=userinfo.get("name"),
            given_name=userinfo.get("given_name"),
            family_name=userinfo.get("family_name"),
            nickname=userinfo.get("nickname"),
            email=userinfo.get("email"),
            email_verified=userinfo.get("email_verified"),
            picture=userinfo.get("picture"),
            updated_at=userinfo.get("updated_at"),
        )


class UserInfo(BaseModel):
    """Class to store info about the user."""

    id: str  # token.sub
    org_id: str  # token.org_id
    org_name: str | None  # token.org_name
    role: str  # token.CLAIM_ROLE
    token: TokenInfo
    profile: UserProfile | None = None

    @classmethod
    def from_claims_and_userinfo(cls, claims: dict[str, Any], userinfo: dict[str, Any] | None = None) -> "UserInfo":
        """Create UserInfo from JWT claims and optional auth0 userinfo.

        Args:
            claims (dict[str, Any]): JWT token claims dictionary.
            userinfo (dict[str, Any] | None): Optional user information dictionary from auth0.

        Returns:
            UserInfo: User information extracted from claims.
        """
        return cls(
            id=claims["sub"],
            org_id=claims["org_id"],
            org_name=claims.get("org_name"),
            role=claims[CLAIM_ROLE],
            token=TokenInfo.from_claims(claims),
            profile=UserProfile.from_userinfo(userinfo) if userinfo else None,
        )


# Services derived from BaseService and exported by modules via their __init__.py are automatically registered
# with the system module, enabling for dynamic discovery of health, info and further functionality.
class Service(BaseService):
    """Service of the application module."""

    _settings: Settings

    def __init__(self) -> None:
        """Initialize service."""
        super().__init__(Settings)  # automatically loads and validates the settings

    def info(self, mask_secrets: bool = True) -> dict[str, Any]:
        """Determine info of this service.

        Args:
            mask_secrets (bool): Whether to mask sensitive information in the output.

        Returns:
            dict[str,Any]: The info of this service.
        """
        user_info = self.get_user_info(relogin=mask_secrets)
        return {
            "settings": self._settings.model_dump(context={UNHIDE_SENSITIVE_INFO: not mask_secrets}),
            "userinfo": user_info.model_dump() if user_info else None,
        }

    def _determine_api_public_health(self) -> Health:
        """Determine healthiness and reachability of Aignostics Platform API.

        - Checks if health endpoint is reachable and returns 200 OK
        - Uses urllib3 for a direct connection check without authentication

        Returns:
            Health: The healthiness of the Aignostics Platform API via basic unauthenticated request.
        """
        try:
            http = urllib3.PoolManager(timeout=urllib3.Timeout(connect=5.0, read=10.0))
            response = http.request(
                method="GET",
                url=f"{self._settings.api_root}/api/v1/health",
                headers={"User-Agent": f"aignostics-python-sdk/{__version__}"},
            )

            if response.status != HTTPStatus.OK:
                logger.error("Aignostics Platform API (public) returned '%s'", response.status)
                return Health(
                    status=Health.Code.DOWN, reason=f"Aignostics Platform API returned status '{response.status}'"
                )
        except Exception as e:
            logger.exception("Issue with Aignostics Platform API")
            return Health(status=Health.Code.DOWN, reason=f"Issue with Aignostics Platform API: '{e}'")

        return Health(status=Health.Code.UP)

    def _determine_api_authenticated_health(self) -> Health:
        """Determine healthiness and reachability of Aignostics Platform API via authenticated API client.

        - Checks if health endpoint is reachable and returns 200 OK

        Returns:
            Health: The healthiness of the Aignostics Platform API when trying to reach via authenticated API client.
        """
        try:
            client = Client()
            api_client = client.get_api_client(cache_token=True).api_client
            response = api_client.call_api(
                url=self._settings.api_root + "/api/v1/health",
                method="GET",
            )
            if response.status != HTTPStatus.OK:
                logger.error("Aignostics Platform API (authenticated) returned '%s'", response.status)
                return Health(status=Health.Code.DOWN, reason=f"Aignostics Platform API returned '{response.status}'")
        except Exception as e:
            logger.exception("Issue with Aignostics Platform API")
            return Health(status=Health.Code.DOWN, reason=f"Issue with Aignostics Platform API: '{e}'")
        return Health(status=Health.Code.UP)

    def health(self) -> Health:
        """Determine health of this service.

        Returns:
            Health: The health of the service.
        """
        return Health(
            status=Health.Code.UP,
            components={
                "api_public": self._determine_api_public_health(),
                "api_authenticated": self._determine_api_authenticated_health(),
            },
        )

    @staticmethod
    def logout() -> bool:
        """Logout if authenticated.

        Deletes the cached authentication token if existing.

        Returns:
            bool: True if successfully logged out, False if not logged in.
        """
        logger.debug("Logging out...")
        rtn = remove_cached_token()
        logger.debug("Logout successful: %s", rtn)
        return rtn

    @staticmethod
    def login(relogin: bool = False) -> bool:
        """Login.

        Args:
            relogin (bool): If True, forces a re-login even if a token is cached.

        Returns:
            bool: True if successfully logged in, False if login failed
        """
        if relogin:
            Service.logout()
        try:
            _ = get_token(use_cache=True)
            return True
        except RuntimeError as e:
            message = f"Error during login: {e!s}"
            logger.exception(message)
            return False

    @staticmethod
    def get_user_info(relogin: bool = False) -> UserInfo | None:
        """Get user information from authentication token.

        Args:
            relogin (bool): If True, forces a re-login even if a token is cached.

        Returns:
            UserInfo | None: User information if successfully authenticated, None if login failed.
        """
        if relogin:
            Service.logout()
        try:
            token = get_token(use_cache=True)
            claims = verify_and_decode_token(token)
            info = None
            try:
                info = userinfo(token)
            except RuntimeError as e:
                message = f"Error retrieving user info: {e!s}"
                logger.exception(message)
            return UserInfo.from_claims_and_userinfo(claims, info)
        except RuntimeError as e:
            message = f"Error during login: {e!s}"
            logger.exception(message)
            return None



from typing import Any

from nicegui import app, background_tasks, context, ui  # noq
from nicegui import run as nicegui_run

from aignostics.gui import frame
from aignostics.utils import get_logger

from .._service import Service  # noqa: TID252
from ._utils import application_id_to_icon, run_status_to_icon_and_color

logger = get_logger(__name__)

BORDERED_SEPARATOR = "bordered separator"
RUNS_LIMIT = 100
STORAGE_TAB_RUNS_COMPLETED_ONLY = "runs_completed_only"

service = Service()


async def _frame(  # noqa: C901, PLR0913, PLR0915, PLR0917
    navigation_title: str,
    navigation_icon: str | None = None,
    navigation_icon_color: str | None = None,
    navigation_icon_tooltip: str | None = None,
    left_sidebar: bool = False,
    args: dict[str, Any] | None = None,
) -> None:
    if args is None:
        args = {}
    with frame(  # noqa: PLR1702
        navigation_title=navigation_title,
        navigation_icon=navigation_icon,
        navigation_icon_color=navigation_icon_color,
        navigation_icon_tooltip=navigation_icon_tooltip,
        left_sidebar=left_sidebar,
    ):
        try:
            with ui.list().props(BORDERED_SEPARATOR).classes("full-width"):
                ui.item_label("Applications").props("header")
                ui.separator()
                for application in service.applications():
                    with (
                        ui.item(
                            on_click=lambda app_id=application.application_id: ui.navigate.to(f"/application/{app_id}")
                        )
                        .mark(f"SIDEBAR_APPLICATION:{application.application_id}")
                        .props("clickable")
                    ):
                        with (
                            ui.item_section().props("avatar"),
                            ui.icon(application_id_to_icon(application.application_id), color="primary"),
                        ):
                            ui.tooltip(application.application_id)
                        with ui.item_section():
                            ui.label(f"{application.name}").tailwind.font_weight(
                                "bold"
                                if context.client.page.path == "/application/{application_id}"
                                and args
                                and args.get("application_id") == application.application_id
                                else "normal"
                            )
        except Exception as e:  # noqa: BLE001
            ui.label(f"Failed to list applications: {e!s}").mark("LABEL_ERROR")

        async def application_runs_load_and_render(runs_column: ui.column, completed_only: bool = False) -> None:
            with runs_column:
                try:
                    runs = await nicegui_run.cpu_bound(
                        Service.application_runs_static, limit=RUNS_LIMIT, completed_only=completed_only
                    )
                    if runs is None:
                        message = "nicegui_run.cpu_bound(Service.application_runs_static) returned None"  # type: ignore[unreachable]
                        logger.error(message)
                        raise RuntimeError(message)  # noqa: TRY301
                    runs_column.clear()
                    for index, run_data in enumerate(runs):
                        with (
                            ui.item(
                                on_click=lambda run_id=run_data["application_run_id"]: ui.navigate.to(
                                    f"/application/run/{run_id}"
                                )
                            )
                            .props("clickable")
                            .classes("w-full")
                            .mark(f"SIDEBAR_RUN_ITEM:{index}")
                        ):
                            with ui.item_section().props("avatar"):
                                icon, color = run_status_to_icon_and_color(run_data["status"])
                                with ui.icon(icon, color=color):
                                    ui.tooltip(
                                        f"Run {run_data['application_run_id']}, "
                                        f"status {run_data['status'].value.upper()}"
                                    )
                            with ui.item_section():
                                ui.label(f"{run_data['application_version_id']}").tailwind.font_weight(
                                    "bold"
                                    if context.client.page.path == "/application/run/{application_run_id}"
                                    and args
                                    and args.get("application_run_id") == run_data["application_run_id"]
                                    else "normal"
                                )
                                ui.label(
                                    f"triggered on {run_data['triggered_at'].astimezone().strftime('%m-%d %H:%M')}"
                                )
                    if not runs:
                        with ui.item():
                            with ui.item_section().props("avatar"):
                                ui.icon("info")
                            with ui.item_section():
                                ui.label("You did not yet create a run.")
                except Exception:
                    runs_column.clear()
                    with ui.item():
                        with ui.item_section().props("avatar"):
                            ui.icon("error")
                        with ui.item_section():
                            ui.label("Failed to load application runs.")
                    logger.exception("Failed to load application runs")

        @ui.refreshable
        async def _runs_list() -> None:
            with ui.column().classes("full-width justify-center") as runs_column:
                with ui.row().classes("w-full justify-center"):
                    ui.spinner(size="lg").classes("m-5")
                await ui.context.client.connected()
                background_tasks.create_lazy(
                    coroutine=application_runs_load_and_render(
                        runs_column=runs_column,
                        completed_only=app.storage.tab.get(STORAGE_TAB_RUNS_COMPLETED_ONLY, False),
                    ),
                    name="_runs_list",
                )

        class RunFilterButton(ui.icon):
            def __init__(self, *args, **kwargs) -> None:  # type: ignore[no-untyped-def]
                super().__init__(*args, **kwargs)
                self._state = app.storage.tab.get(STORAGE_TAB_RUNS_COMPLETED_ONLY, False)
                self.on("click", self.toggle)

            def toggle(self) -> None:
                self._state = not self._state
                app.storage.tab[STORAGE_TAB_RUNS_COMPLETED_ONLY] = self._state
                self.update()
                _runs_list.refresh()

            def update(self) -> None:
                self.props(f"color={'positive' if self._state else 'grey'}")
                super().update()

            def is_active(self) -> bool:
                return bool(self._state)

        try:
            with ui.list().props(BORDERED_SEPARATOR).classes("full-width"):
                with ui.row(align_items="center").classes("justify-center"):
                    ui.item_label("Runs").props("header")
                    ui.space()
                    await ui.context.client.connected()
                    with RunFilterButton("done_all", size="sm").classes("mr-3").mark("BUTTON_RUNS_FILTER_COMPLETED"):
                        ui.tooltip("Show completed runs only")
                ui.separator()
                await _runs_list()
        except Exception as e:  # noqa: BLE001
            ui.label(f"Failed to list application runs: {e!s}").mark("LABEL_ERROR")



"""Application module."""

from ._cli import cli
from ._service import Service

__all__ = [
    "Service",
    "cli",
]

from importlib.util import find_spec

# advertise PageBuilder to enable auto-discovery
if find_spec("nicegui"):
    from ._gui._page_builder import PageBuilder

    __all__ += [
        "PageBuilder",
    ]



"""Utility functions for the application GUI."""

from aignostics.platform import ApplicationRunStatus, ItemStatus


def application_id_to_icon(application_id: str) -> str:
    """Convert application ID to icon.

    Args:
        application_id (str): The application ID.

    Returns:
        str: The icon name.
    """
    match application_id:
        case "he-tme":
            return "biotech"
        case "test-app":
            return "construction"
    return "bug_report"


def run_status_to_icon_and_color(run_status: str) -> tuple[str, str]:  # noqa: PLR0911
    """Convert run status to icon amd color.

    Args:
        run_status (str): The run status.

    Returns:
        tuple[str, str]: The icon name and color.
    """
    match run_status:
        case ApplicationRunStatus.RUNNING:
            return "directions_run", "info"
        case ApplicationRunStatus.CANCELED_USER:
            return "cancel", "warning"
        case ApplicationRunStatus.CANCELED_SYSTEM:
            return "sync_problem", "negative"
        case ApplicationRunStatus.COMPLETED:
            return "done_all", "positive"
        case ApplicationRunStatus.COMPLETED_WITH_ERROR:
            return "error", "negative"
        case ApplicationRunStatus.RECEIVED:
            return "call_received", "info"
        case ApplicationRunStatus.REJECTED:
            return "hand_gesture_off", "negative"
        case ApplicationRunStatus.RUNNING:
            return "directions_run", "info"
        case ApplicationRunStatus.SCHEDULED:
            return "schedule", "info"
    return "bug_report", "negative"


def run_item_status_to_icon_and_color(run_status: str) -> tuple[str, str]:  # noqa: PLR0911
    """Convert run item status to icon.

    Args:
        run_status (str): The run item status.

    Returns:
        tuple[str, str]: The icon name and color.
    """
    match run_status:
        case ItemStatus.PENDING:
            return "pending", "info"
        case ItemStatus.CANCELED_USER:
            return "cancel", "warning"
        case ItemStatus.CANCELED_SYSTEM:
            return "sync_problem", "negative"
        case ItemStatus.ERROR_USER:
            return "hand_gesture_off", "negative"
        case ItemStatus.ERROR_SYSTEM:
            return "error", "negative"
        case ItemStatus.SUCCEEDED:
            return "check", "positive"
    return "bug_report", "negative"


def mime_type_to_icon(mime_type: str) -> str:
    """Convert mime type to icon.

    Args:
        mime_type (str): The mime type.

    Returns:
        str: The icon name.
    """
    match mime_type:
        case "image/tiff":
            return "image"
        case "application/dicom":
            return "image"
        case "text/csv":
            return "table_rows"
        case "application/geo+json":
            return "place"
        case "application/json":
            return "data_object"
    return "bug_report"



"""CLI of bucket module."""

import datetime
import json
from pathlib import Path
from typing import Annotated

import humanize
import typer

from aignostics.utils import console, get_logger

from ._service import Service

MESSAGE_NOT_YET_IMPLEMENTED = "NOT YET IMPLEMENTED"

logger = get_logger(__name__)


cli = typer.Typer(
    name="bucket",
    help="Operations on cloud bucket on Aignostics Platform.",
)


@cli.command()
def upload(
    source: Annotated[
        Path,
        typer.Argument(
            help="Source file or directory to upload",
            exists=True,
            file_okay=True,
            dir_okay=True,
            writable=False,
            readable=True,
            resolve_path=False,
        ),
    ],
    destination_prefix: Annotated[
        str,
        typer.Option(
            help="Destination layout. Supports {username}, {timestamp}. "
            'E.g. you might want to use "{username}/myproject/"'
        ),
    ] = "{username}",
) -> None:
    """Upload file or directory to bucket in Aignostics platform."""
    import psutil  # noqa: PLC0415
    from rich.progress import (  # noqa: PLC0415
        BarColumn,
        FileSizeColumn,
        Progress,
        TaskProgressColumn,
        TextColumn,
        TimeRemainingColumn,
        TotalFileSizeColumn,
        TransferSpeedColumn,
    )

    console.print(f"Uploading {source} to bucket...")

    total_bytes = 0
    files_count = 0

    if source.is_file():
        total_bytes = source.stat().st_size
        files_count = 1
    else:
        for file_path in source.glob("**/*"):
            if file_path.is_file():
                total_bytes += file_path.stat().st_size
                files_count += 1

    console.print(f"Found {files_count} files with total size of {humanize.naturalsize(total_bytes)}")

    username = psutil.Process().username()
    timestamp = datetime.datetime.now(tz=datetime.UTC).strftime("%Y%m%d_%H%M%S")
    base_prefix = destination_prefix.format(username=username, timestamp=timestamp)
    base_prefix = base_prefix.strip("/")

    with Progress(
        TextColumn(
            f"[progress.description]Uploading from {source.name} to "
            f"{Service().get_bucket_protocol()}:/{Service().get_bucket_name()}/{base_prefix}"
        ),
        BarColumn(),
        TaskProgressColumn(),
        TimeRemainingColumn(),
        FileSizeColumn(),
        TotalFileSizeColumn(),
        TransferSpeedColumn(),
        TextColumn("[progress.description]{task.description}"),
    ) as progress:
        task = progress.add_task(f"Uploading to {base_prefix}/...", total=total_bytes)

        def update_progress(bytes_uploaded: int, file: Path) -> None:
            relpath = file.relative_to(source)
            progress.update(task, advance=bytes_uploaded, description=f"{relpath}")

        results = Service().upload(source, base_prefix, update_progress)

    if results["success"]:
        console.print(f"[green]Successfully uploaded {len(results['success'])} files:[/green]")
        for key in results["success"]:
            console.print(f"  [green]- {key}[/green]")

    if results["failed"]:
        console.print(f"[red]Failed to upload {len(results['failed'])} files:[/red]")
        for key in results["failed"]:
            console.print(f"  [red]- {key}[/red]")

    if not results["failed"]:
        console.print("[green]All files uploaded successfully![/green]")


@cli.command()
def ls(
    detail: Annotated[bool, typer.Option(help="Show details")] = False,
) -> None:
    """List objects in bucket on Aignostics Platform."""
    console.print_json(json=json.dumps(Service().ls(detail=detail), default=str))


@cli.command()
def find(
    detail: Annotated[bool, typer.Option(help="Show details")] = False,
) -> None:
    """Find objects in bucket on Aignostics Platform."""
    console.print_json(json=json.dumps(Service().find(detail=detail), default=str))


@cli.command()
def delete(
    key: Annotated[str, typer.Argument(help="key of object in object")],
) -> None:
    """Find objects in bucket on Aignostics Platform."""
    deleted = Service().delete_objects([key])
    if deleted:
        console.print(f"Deleted object with key '{key}'")
    else:
        console.print(f"Object with key '{key}' not found")


@cli.command()
def purge() -> None:
    """Purge all objects in bucket on Aignostics Platform."""
    console.print(MESSAGE_NOT_YET_IMPLEMENTED)



"""Notebook service running Marimo."""

import atexit
import re
import sys
from subprocess import PIPE, STDOUT, Popen
from threading import Event, Thread
from typing import Any

from aignostics.constants import NOTEBOOK_DEFAULT
from aignostics.utils import BaseService, Health, get_logger, get_user_data_directory

logger = get_logger(__name__)

MARIMO_SERVER_STARTUP_TIMEOUT = 30


class _Runner:
    """Runner class of the Marimo module."""

    _marimo_server: Popen[str] | None = None
    _monitor_thread: Thread | None = None
    _output: str = ""
    _server_url: str | None = None
    _server_ready: Event = Event()
    _started = False

    def __init__(self) -> None:
        atexit.register(self.stop)

    def health(self) -> Health:
        """Determine health of hello service.

        Returns:
            Health: The health of the service.
        """
        if not self._started:
            return Health(status=Health.Code.UP)
        return Health(
            status=Health.Code.UP,
            components={
                "marimo_server": Health(
                    status=Health.Code.UP if self.is_marimo_server_running() else Health.Code.DOWN,
                    reason=None if self.is_marimo_server_running() else "Marimo server is not running.",
                ),
                "monitor_thread": Health(
                    status=Health.Code.UP if self.is_monitor_thread_alive() else Health.Code.DOWN,
                    reason=None if self.is_monitor_thread_alive() else "Monitor thread is not running.",
                ),
            },
        )

    def start(self, timeout: int = MARIMO_SERVER_STARTUP_TIMEOUT) -> str:
        """Start the Marimo server.

        Args:
            timeout (int): Maximum time to wait for the server to start and URL to be detected.

        Returns:
            str: The URL of the started Marimo server.

        Raises:
            RuntimeError: If the Marimo server fails to start or if the URL isn't detected within given timeout.
        """
        self._started = True
        if self.is_marimo_server_running():
            logger.warning("Marimo server is already running.")
            if self._server_url is not None:
                return self._server_url

            message = "Server is running but URL is not set - this is unexpected."
            logger.error(message)
            raise RuntimeError(message)

        directory = get_user_data_directory("notebooks")
        notebook_path = directory / "notebook.py"
        if not notebook_path.exists():
            logger.debug("Copying notebook to user data directory '%s'...", notebook_path)
            notebook_path.write_bytes(NOTEBOOK_DEFAULT.read_bytes())

        # Reset server state
        self._server_url = None
        self._server_ready.clear()

        logger.debug("Starting Marimo server with notebook at '%s'...", notebook_path)
        self._marimo_server = Popen(  # noqa: S603
            [
                sys.executable,
                "-m",
                "marimo",
                "edit",
                "--headless",
                "--skip-update-check",
                "--no-sandbox",
                "--no-token",
                str(notebook_path.resolve()),
            ],
            stdout=PIPE,
            stderr=STDOUT,
            text=True,
            encoding="utf-8",
            errors="replace",
            bufsize=1,
        )

        # Start a thread to monitor the subprocess output
        self._monitor_thread = Thread(target=self._capture_output, args=(self._marimo_server,), daemon=True)
        self._monitor_thread.start()

        # Wait up to timeout seconds for the server URL to be detected
        if not self._server_ready.wait(timeout=timeout):
            self.stop()  # Kill the process if it didn't start properly
            message = f"Marimo server didn't start within '{timeout}' seconds (URL not detected)."
            logger.error(message)
            raise RuntimeError(message)

        # At this point, self._server_url should be set by thread
        if self._server_url is None:
            self.stop()
            message = "Server URL was not set despite server ready event being triggered."
            logger.error(message)
            raise RuntimeError(message)

        message = f"Marimo server started successfully with URL '{self._server_url}'."  # type: ignore[unreachable]
        logger.info(message)
        return self._server_url

    def _capture_output(self, process: Popen[str]) -> None:
        """Capture stdout of the subprocess and detect when server is ready.

        Args:
            process (Popen): The subprocess to capture stdout from.
        """
        captured_line = ""
        url_pattern = re.compile(r"URL:\s+((?:http|https)://[^\s]{1,100})")

        if process.stdout is None:
            logger.warning("Cannot capture stdout")
            return

        # Buffer for collecting output
        self._output = ""

        while process.poll() is None:
            char = process.stdout.read(1)
            if not char:  # End of stream
                break

            self._output += char
            captured_line += char

            if char == "\n":
                logger.debug(captured_line.rstrip())
                url_match = url_pattern.search(captured_line)
                if url_match:
                    self._server_url = url_match.group(1)
                    logger.info("Found URL: '%s'", self._server_url)
                    self._server_ready.set()

                captured_line = ""

        logger.debug("Marimo server process completed.")

    def is_marimo_server_running(self) -> bool:
        """Check if the marimo server is running.

        Returns:
            bool: True if the server is running, False otherwise.
        """
        return self._marimo_server is not None and self._marimo_server.poll() is None

    def is_monitor_thread_alive(self) -> bool:
        """Check if the monitor thread is running.

        Returns:
            bool: True if the thread is running, False otherwise.
        """
        return self._monitor_thread is not None and self._monitor_thread.is_alive()

    def stop(self) -> None:
        """Stop the Marimo server."""
        if self._marimo_server is not None:
            logger.debug("Stopping Marimo server...")
            self._marimo_server.terminate()
            self._marimo_server.wait(2)
            if self._marimo_server.returncode is None:
                logger.debug("Marimo server did not terminate in time, killing it...")
                self._marimo_server.kill()
            self._marimo_server = None
            logger.info("Marimo server stopped.")
        else:
            logger.debug("Marimo server is not running.")
        if self._monitor_thread is not None:
            self._monitor_thread.join()
            self._monitor_thread = None
            logger.info("Monitor thread stopped.")
        else:
            logger.debug("Monitor thread is not running.")
        logger.info("Service stopped.")


# Singleton instance of Runner
runner: _Runner | None = None


# Lazy init the runner
def _get_runner() -> _Runner:
    """Get the singleton runner.

    Returns:
        Service: The service instance.
    """
    global runner  # noqa: PLW0603
    if runner is None:
        runner = _Runner()
    return runner


# Facade to the Marimo runner
class Service(BaseService):
    """Service of the Marimo module."""

    def info(self, mask_secrets: bool = True) -> dict[str, Any]:  # noqa: ARG002, PLR6301
        """Determine info of this service.

        Args:
            mask_secrets (bool): Whether to mask sensitive information in the output.

        Returns:
            dict[str,Any]: The info of this service.
        """
        return {}

    def health(self) -> Health:  # noqa: PLR6301
        """Determine health of hello service.

        Returns:
            Health: The health of the service.
        """
        return _get_runner().health()

    def start(self) -> str:  # noqa: PLR6301
        """Start the Marimo server.

        Returns:
            str: The URL of the started Marimo server.

        Raises:
            RuntimeError: If the Marimo server fails to start or if the URL isn't detected within 10 seconds.
        """
        return _get_runner().start()

    def is_marimo_server_running(self) -> bool:  # noqa: PLR6301
        """Check if the marimo server is running.

        Returns:
            bool: True if the server is running, False otherwise.
        """
        return _get_runner().is_marimo_server_running()

    def is_monitor_thread_alive(self) -> bool:  # noqa: PLR6301
        """Check if the monitor thread is running.

        Returns:
            bool: True if the thread is running, False otherwise.
        """
        return _get_runner().is_monitor_thread_alive()

    def stop(self) -> None:  # noqa: PLR6301
        """Stop the Marimo server."""
        return _get_runner().stop()



"""GUI of application module including homepage of Aignostics Launchpad."""

from importlib.util import find_spec

from nicegui import Client, app, ui  # noq

from aignostics.utils import get_logger

from ._frame import _frame

logger = get_logger(__name__)


async def _page_index(client: Client) -> None:
    """Homepage of Applications."""
    client.content.classes(remove="nicegui-content")
    client.content.classes(add="pl-5 pt-5")

    await _frame("Analyze your Whole Slide Images with AI", left_sidebar=True)

    with ui.row().classes("p-4 pt-2 pr-0"), ui.column():
        await ui.context.client.connected()
        ui.label("Welcome to the Aignostics Launchpad").bind_text_from(
            app.storage.tab,
            "user_info",
            lambda user_info: (
                f"Welcome {user_info.profile.given_name if user_info.profile else ''} to the Aignostics Launchpad"
            ),
        ).classes("text-4xl mb-2")
        ui.label(
            "Using the Launchpad, you can run Aignostics Atlas applications on your whole slide images, "
            "monitor progress, and download results for analysis."
        ).classes("text-2xl")
        with ui.row().classes("w-full h-screen flex"):
            with ui.column().classes("flex-1 pr-0"):
                ui.label("Have your slides ready to go?").classes("text-xl")
                ui.label(
                    "Select the relevant application from the left sidebar to upload your images to get started."
                ).classes("text")
                ui.label("Looking to test with public data first?").classes("text-xl")
                ui.label(
                    'Open ☰ menu and click "Download Datasets" to select a dataset '
                    "from Image Data Commons by the National Cancer Institute to test with."
                ).classes("text")
                ui.label("Already submitted your slides?").classes("text-xl")
                ui.label(
                    "Select your run on the left to monitor progress, cancel while pending, or download results."
                ).classes("text")

                if find_spec("ijson"):
                    from aignostics.qupath import Service as QuPathService  # noqa: PLC0415

                    ui.label("Visualize results in QuPath Microscopy Viewer?").classes("text-xl")
                    if QuPathService.is_installed():
                        ui.label('Select a completed run and click "QuPath" to visualize results.').classes("text")
                    else:
                        ui.label(
                            "To visualize results with one click the QuPath application must be installed. "
                            'Select "QuPath extension" in the ☰ menu to install.'
                        ).classes("text")

                if find_spec("marimo"):
                    ui.label("Analyze results in Python Notebooks?").classes("text-xl")
                    ui.label('On completed runs click "Marimo" to directly open the notebook.').classes("text")

            with (
                ui.carousel(animated=True, arrows=True, navigation=True)
                .classes("flex-1 h-full m-0 p-0 self-end bg-[#423D6B] ")
                .props("infinite autoplay=1000 control-color=transparent")
            ):
                for i in range(1, 5):  # Loop from 1 to 4
                    with ui.carousel_slide().classes("p-0 m-0"):
                        ui.image(f"/application_assets/home-card-{i}.png").classes("w-full h-auto object-contain")
